"""Administration form for general Review Board settings."""

from collections import OrderedDict
from urllib.parse import urlparse

from django import forms
from django.conf import settings
from django.contrib.sites.models import Site
from django.utils.module_loading import import_string
from django.utils.translation import gettext, gettext_lazy as _
from djblets.cache.backend_compat import normalize_cache_backend
from djblets.cache.forwarding_backend import DEFAULT_FORWARD_CACHE_ALIAS
from djblets.forms.fields import TimeZoneField
from djblets.siteconfig.forms import SiteSettingsForm

from reviewboard.admin.siteconfig import load_site_config


class BaseCacheSettingsForm(forms.Form):
    """Base class for a cache backend settings form.

    Version Added:
        4.0
    """

    def load(self, cache_backend_settings):
        """Load the cache backend settings into the form.

        Args:
            cache_backend_settings (dict):
                Settings for the cache backend.
        """
        pass

    def build_cache_backend_settings(self):
        """Build new cache backend settings from the form data.

        Returns:
            dict:
            The new cache backend settings.
        """
        raise NotImplementedError


class FileCacheSettingsForm(BaseCacheSettingsForm):
    """Settings for the file-based cache backend.

    This backend is recommended only for simple test setups and development.
    """

    cache_path = forms.CharField(
        label=_('Cache Path'),
        help_text=_(
            'The file location for the cache. This must be writable by '
            'the web server.'
        ),
        required=True,
        widget=forms.TextInput(attrs={'size': '50'}),
        error_messages={
            'required': 'A valid cache path must be provided.'
        })

    def load(self, cache_backend_settings):
        """Load the cache backend settings into the form.

        Args:
            cache_backend_settings (dict):
                Settings for the cache backend.
        """
        cache_locations = cache_backend_settings['LOCATION']

        if not isinstance(cache_locations, list):
            cache_locations = [cache_locations]

        self.fields['cache_path'].initial = ';'.join(cache_locations)

    def build_cache_backend_settings(self):
        """Build new cache backend settings from the form data.

        Returns:
            dict:
            The new cache backend settings.
        """
        return {
            'LOCATION': self.cleaned_data['cache_path'],
        }

    def clean_cache_path(self):
        """Clean the cache_path field.

        This will strip whitespace around the value and return it.

        Returns:
            unicode:
            The cache path, with whitespace stripped.
        """
        return self.cleaned_data['cache_path'].strip()

    class Meta:
        title = _('Local File Cache Settings')


class LocalMemoryCacheSettingsForm(BaseCacheSettingsForm):
    """Settings for the local memory cache backend.

    This is only available for development setups and cannot be used in
    production.
    """

    def build_cache_backend_settings(self):
        """Build new cache backend settings from the form data.

        Returns:
            dict:
            The new cache backend settings.
        """
        # We want to specify a "reviewboard" location to keep items
        # separate from those in other caches.
        return {
            'LOCATION': 'reviewboard'
        }

    class Meta:
        title = _('Local Memory Cache Settings')


class MemcachedSettingsForm(BaseCacheSettingsForm):
    """Settings for the memcached backend.

    This is the recommended caching backend.
    """

    cache_host = forms.CharField(
        label=_('Cache Hosts'),
        help_text=_('The host or hosts used for the cache, in hostname:port '
                    'form. Multiple hosts can be specified by separating '
                    'them with a semicolon (;).'),
        required=True,
        widget=forms.TextInput(attrs={'size': '50'}),
        error_messages={
            'required': 'A valid cache host must be provided.'
        })

    def load(self, cache_backend_settings):
        """Load the cache backend settings into the form.

        Args:
            cache_backend_settings (dict):
                Settings for the cache backend.
        """
        cache_locations = cache_backend_settings['LOCATION']

        if not isinstance(cache_locations, list):
            cache_locations = [cache_locations]

        self.fields['cache_host'].initial = ';'.join(cache_locations)

    def build_cache_backend_settings(self):
        """Build new cache backend settings from the form data.

        Returns:
            dict:
            The new cache backend settings.
        """
        return {
            'LOCATION': self.cleaned_data['cache_host'].split(';'),
        }

    def clean_cache_host(self):
        """Clean the cache_host field.

        This will strip whitespace around the value and return it.

        Returns:
            unicode:
            The cache host, with whitespace stripped.
        """
        return self.cleaned_data['cache_host'].strip()

    class Meta:
        title = _('Memcached Settings')


class GeneralSettingsForm(SiteSettingsForm):
    """General settings for Review Board."""

    _cache_backends = OrderedDict([
        ('locmem', {
            'name': _('Local memory cache (developer-only)'),
            'available': not settings.PRODUCTION,
            'backend_cls_path':
                'django.core.cache.backends.locmem.LocMemCache',
            'form_cls': LocalMemoryCacheSettingsForm,
        }),
        ('memcached', {
            'name': _('Memcached (recommended)'),
            'backend_cls_path':
                'django.core.cache.backends.memcached.PyMemcacheCache',
            'legacy_backend_cls_paths': [
                'django.core.cache.backends.memcached.CacheClass',
                'django.core.cache.backends.memcached.MemcachedCache',
            ],
            'form_cls': MemcachedSettingsForm,

        }),
        ('file', {
            'name': _('Local file cache'),
            'backend_cls_path':
                'django.core.cache.backends.filebased.FileBasedCache',
            'form_cls': FileCacheSettingsForm,
        }),
    ])

    CACHE_VALIDATION_KEY = '__rb-cache-validation__'
    CACHE_VALIDATION_VALUE = 12345

    company = forms.CharField(
        label=_('Company/Organization'),
        help_text=_('The optional name of your company or organization. '
                    'This will be displayed on your support page.'),
        required=False,
        widget=forms.TextInput(attrs={'size': '30'}))

    server = forms.CharField(
        label=_('Server'),
        help_text=_('The URL of this Review Board server. This should not '
                    'contain the subdirectory Review Board is installed in.'),
        widget=forms.TextInput(attrs={'size': '30'}))

    site_read_only = forms.BooleanField(
        label=_('Enable read-only mode'),
        help_text=_('Prevent non-superusers from making any changes to '
                    'Review Board.'),
        required=False)

    read_only_message = forms.CharField(
        label=_('Read-only message'),
        help_text=_('A custom message displayed when the site is in '
                    'read-only mode.'),
        required=False,
        widget=forms.TextInput(attrs={'size': '30'}))

    site_media_url = forms.CharField(
        label=_('Media URL'),
        help_text=(_('The URL to the media files. Set to '
                     '<code>%smedia/</code> to use the default media path on '
                     'this server.')
                   % settings.SITE_ROOT),
        required=True,
        widget=forms.TextInput(attrs={'size': '30'}))

    site_static_url = forms.CharField(
        label=_('Static URL'),
        help_text=(_('The URL to the static files, such as JavaScript files, '
                     'CSS files, and images that are bundled with Review '
                     'Board or third-party extensions. Set to '
                     '<code>%sstatic/</code> to use the default static path '
                     'on this server.')
                   % settings.SITE_ROOT),
        required=True,
        widget=forms.TextInput(attrs={'size': '30'}))

    site_admin_name = forms.CharField(
        label=_('Administrator Name'),
        required=True,
        widget=forms.TextInput(attrs={'size': '30'}))
    site_admin_email = forms.EmailField(
        label=_('Administrator E-Mail'),
        required=True,
        widget=forms.TextInput(attrs={'size': '30'}))

    locale_timezone = TimeZoneField(
        label=_('Time Zone'),
        required=True,
        help_text=_('The time zone used for all dates on this server.'))

    cache_type = forms.ChoiceField(
        label=_('Cache Backend'),
        help_text=_('The type of server-side caching to use.'),
        required=True,
        widget=forms.Select(attrs={
            'data-subform-group': 'cache-backend',
        }))

    def __init__(self, siteconfig, data=None, files=None, *args, **kwargs):
        """Initialize the settings form.

        Args:
            siteconfig (djblets.siteconfig.models.SiteConfiguration):
                The site configuration being changed.

            data (dict, optional):
                The submitted form data.

            files (dict, optional):
                The uploaded file data.

            *args (tuple):
                Additional positional arguments for the form.

            **kwargs (dict):
                Additional keyword arguments for the form.
        """
        self.cache_backend_forms = {
            backend_id: backend_info['form_cls'](data=data, files=files)
            for backend_id, backend_info in self._cache_backends.items()
            if backend_info.get('available', True)
        }

        super(GeneralSettingsForm, self).__init__(siteconfig,
                                                  data=data,
                                                  files=files,
                                                  *args, **kwargs)

    def load(self):
        """Load settings from the form.

        This will populate initial fields based on the site configuration.
        It takes care to transition legacy (<= Review Board 1.7) cache
        backends, if still used in production, to a modern configuration.
        """
        domain_method = self.siteconfig.get('site_domain_method')
        site = Site.objects.get_current()

        # Load the rest of the settings from the form.
        super(GeneralSettingsForm, self).load()

        # Load the cache settings.
        cache_backend_info = self.siteconfig.get('cache_backend')
        cache_backend = (
            normalize_cache_backend(cache_backend_info,
                                    DEFAULT_FORWARD_CACHE_ALIAS) or
            normalize_cache_backend(cache_backend_info))

        cache_backend_path = cache_backend['BACKEND']
        cache_type = 'custom'

        for _cache_type, backend_info in self._cache_backends.items():
            if (cache_backend_path == backend_info['backend_cls_path'] or
                cache_backend_path in backend_info.get(
                    'legacy_backend_cls_paths', [])):
                cache_type = _cache_type
                break

        cache_type_choices = [
            (backend_id, backend_info['name'])
            for backend_id, backend_info in self._cache_backends.items()
            if backend_info.get('available', True)
        ]

        if cache_type == 'custom':
            cache_type_choices.append(('custom', gettext('Custom')))

        cache_type_field = self.fields['cache_type']
        cache_type_field.choices = tuple(cache_type_choices)
        cache_type_field.initial = cache_type

        if cache_type in self.cache_backend_forms:
            self.cache_backend_forms[cache_type].load(cache_backend)

        # This must come after we've loaded the general settings.
        self.fields['server'].initial = '%s://%s' % (domain_method,
                                                     site.domain)

    def save(self):
        """Save the form.

        This will write the new configuration to the database. It will then
        force a site configuration reload.
        """
        server = self.cleaned_data['server']

        if '://' not in server:
            # urlparse doesn't properly handle URLs without a scheme. It
            # believes the domain is actually the path. So we apply a prefix.
            server = 'http://' + server

        url_parts = urlparse(server)
        domain_method = url_parts[0]
        domain_name = url_parts[1]

        if domain_name.endswith('/'):
            domain_name = domain_name[:-1]

        site = Site.objects.get_current()

        if site.domain != domain_name:
            site.domain = domain_name
            site.save(update_fields=['domain'])

        self.siteconfig.set('site_domain_method', domain_method)

        cache_type = self.cleaned_data['cache_type']

        if cache_type != 'custom':
            cache_backend_info = self._cache_backends[cache_type]
            cache_form = self.cache_backend_forms[cache_type]
            cache_backend_settings = cache_form.build_cache_backend_settings()

            self.siteconfig.set('cache_backend', {
                DEFAULT_FORWARD_CACHE_ALIAS: dict({
                    'BACKEND': cache_backend_info['backend_cls_path'],
                }, **cache_backend_settings),
            })

        super(GeneralSettingsForm, self).save()

        # Reload any important changes into the Django settings.
        load_site_config()

    def is_valid(self):
        """Return whether the form is valid.

        This will check that this form, and the form for any selected cache
        backend, is valid.

        Returns:
            bool:
            ``True`` if all form fields are valid. ``False`` if any are
            invalid.
        """
        if not super(GeneralSettingsForm, self).is_valid():
            return False

        cache_form = self.cache_backend_forms.get(
            self.cleaned_data['cache_type'])

        return cache_form is None or cache_form.is_valid()

    def clean(self):
        """Clean and validate the form fields.

        This is called after all individual fields are validated. It does
        the remaining work of checking to make sure the resulting configuration
        is valid.

        Returns:
            dict:
            The cleaned form data.
        """
        cleaned_data = super(GeneralSettingsForm, self).clean()

        if 'cache_type' not in self.errors:
            cache_type = cleaned_data['cache_type']
            cache_backend_info = self._cache_backends.get(cache_type)
            cache_form = self.cache_backend_forms.get(cache_type)

            if (cache_backend_info is not None and
                cache_form is not None and
                cache_form.is_valid()):
                cache_backend = None

                try:
                    cache_backend_settings = \
                        cache_form.build_cache_backend_settings()

                    cache_cls = import_string(
                        cache_backend_info['backend_cls_path'])
                    cache_backend = cache_cls(
                        cache_backend_settings.get('LOCATION'),
                        {})

                    cache_backend.set(self.CACHE_VALIDATION_KEY,
                                      self.CACHE_VALIDATION_VALUE)
                    value = cache_backend.get(self.CACHE_VALIDATION_KEY)
                    cache_backend.delete(self.CACHE_VALIDATION_KEY)

                    if value != self.CACHE_VALIDATION_VALUE:
                        self.errors['cache_type'] = self.error_class([
                            _('Unable to store and retrieve values from this '
                              'caching backend. There may be a problem '
                              'connecting.')
                        ])
                except Exception as e:
                    self.errors['cache_type'] = self.error_class([
                        _('Error with this caching configuration: %s')
                        % e
                    ])

                # If the cache backend is open, try closing it. This may fail,
                # so we want to ignore any failures.
                if cache_backend is not None:
                    try:
                        cache_backend.close()
                    except Exception:
                        pass

        return cleaned_data

    class Meta:
        title = _('General Settings')
        save_blacklist = ('server', 'cache_type')

        subforms = (
            {
                'subforms_attr': 'cache_backend_forms',
                'controller_field': 'cache_type',
            },
        )

        fieldsets = (
            {
                'title': _('Site Settings'),
                'classes': ('wide',),
                'fields': ('company', 'server', 'site_media_url',
                           'site_static_url', 'site_admin_name',
                           'site_admin_email', 'locale_timezone',
                           'site_read_only', 'read_only_message'),
            },
            {
                'title': _('Cache Settings'),
                'classes': ('wide',),
                'fields': ('cache_type',),
            },
        )



"""Add Profile.settings.

Version Added:
    3.0
"""

from django_evolution.mutations import AddField
from djblets.db.fields import JSONField


MUTATIONS = [
    AddField('Profile', 'settings', JSONField, null=True),
]



"""Unit tests for reviewboard.scmtools.managers.ToolManager."""

from django.db.models import Q

from reviewboard.scmtools.models import Tool
from reviewboard.testing import TestCase


class ToolManagerTests(TestCase):
    """Unit tests for reviewboard.scmtools.managers.ToolManager."""

    fixtures = ['test_scmtools']

    def test_get_with_id_caches(self):
        """Testing Tool.objects.get with id= caches"""
        self._test_id_cache_query(id=1)

    def test_get_with_id_exact_caches(self):
        """Testing Tool.objects.get with id__exact= caches"""
        self._test_id_cache_query(id__exact=1)

    def test_get_with_pk_caches(self):
        """Testing Tool.objects.get with pk= caches"""
        self._test_id_cache_query(pk=1)

    def test_get_with_pk_exact_caches(self):
        """Testing Tool.objects.get with pk__exact= caches"""
        self._test_id_cache_query(pk=1)

    def test_get_with_q_id_caches(self):
        """Testing Tool.objects.get with Q(id=) caches"""
        self._test_id_cache_query(Q(id=1))

    def test_get_with_q_pk_caches(self):
        """Testing Tool.objects.get with Q(pk=) caches"""
        self._test_id_cache_query(Q(pk=1))

    def test_get_with_q_id_exact_caches(self):
        """Testing Tool.objects.get with Q(id__exact=) caches"""
        self._test_id_cache_query(Q(id__exact=1))

    def test_get_with_q_pk_exact_caches(self):
        """Testing Tool.objects.get with Q(pk__exact=) caches"""
        self._test_id_cache_query(Q(pk__exact=1))

    def test_get_non_id_lookup(self):
        """Testing Tool.objects.get with non-id/pk lookup"""
        with self.assertNumQueries(1):
            tool1 = Tool.objects.get(name='Git')

        with self.assertNumQueries(2):
            tool2 = Tool.objects.get(name='Git')
            tool3 = Tool.objects.get(name='CVS')

        self.assertIsNot(tool1, tool2)
        self.assertEqual(tool1, tool2)
        self.assertNotEqual(tool1, tool3)

    def _test_id_cache_query(self, *args, **kwargs):
        """Utility function for testing ID-based caching.

        Args:
            *args (tuple):
                Positional arguments to use for the query.

            **kwargs (dict):
                Keyword arguments to use for the query.

        Raises:
            AssertionError:
                An assertion failed.
        """
        Tool.objects.clear_tool_cache()

        with self.assertNumQueries(1):
            tool1 = Tool.objects.get(*args, **kwargs)

        with self.assertNumQueries(0):
            # Further queries for any available ID should reuse the cache.
            tools = [
                Tool.objects.get(id=1),
                Tool.objects.get(id__exact=1),
                Tool.objects.get(pk=1),
                Tool.objects.get(pk__exact=1),
                Tool.objects.get(Q(id=1)),
                Tool.objects.get(Q(id__exact=1)),
                Tool.objects.get(Q(pk=1)),
                Tool.objects.get(Q(pk__exact=1)),
            ]

            tool3 = Tool.objects.get(pk=2)

        for tool in tools:
            self.assertIs(tool1, tool)

        self.assertIsNot(tool1, tool3)



"""Review Board admin panel-specific decorators"""

from functools import wraps

from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.shortcuts import render
from djblets.util.decorators import simple_decorator

from reviewboard.admin.read_only import is_site_read_only_for
from reviewboard.site.urlresolvers import local_site_reverse


def superuser_required(view):
    """Wrap a view so that is only accessible to superusers.

    Unauthenticated users will be redirected to the login page. Logged in users
    without sufficient permissions will be redirected to a page showing a
    permission denied error.

    This is very similar to Django's own
    :py:func:`~django.contrib.admin.views.decorators.staff_member_required`,
    except it checks for superuser status instead of staff status.

    Args:
        view (callable):
            The view to wrap.

    Returns:
        callable:
        The wrapped view.
    """
    @wraps(view)
    def decorated(request, *args, **kwargs):
        if not (request.user.is_active and request.user.is_superuser):
            return render(
                request=request,
                template_name='admin/permission_denied.html',
                context={
                    'user': request.user,
                })

        return view(request, *args, **kwargs)

    return login_required(decorated)


@simple_decorator
def check_read_only(view):
    """Check whether the site is read only.

    If the site is currently in read-only mode, this will redirect to a page
    indicating that state.

    Args:
        view (callable):
            The view to wrap.

    Returns:
        callable:
        The wrapped view.
    """
    def _check_read_only(request, *args, **kwargs):
        if is_site_read_only_for(request.user):
            return HttpResponseRedirect(
                local_site_reverse('read-only', request=request))
        else:
            return view(request, *args, **kwargs)

    return _check_read_only



"""Switch Profile.default_use_rich_text to a BooleanField.

This is required for Django 3.2, which no longer provides the old
:py:class:`django.db.models.BooleanField`.

Version Added:
    5.0
"""

from django.db import models
from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('Profile', 'default_use_rich_text',
                field_type=models.BooleanField,
                null=True),
]



"""Unit tests for reviewboard.reviews.views.NewReviewRequestView."""

from django.contrib.auth.models import User

from djblets.siteconfig.models import SiteConfiguration
from djblets.testing.decorators import add_fixtures

from reviewboard.testing import TestCase


class NewReviewRequestViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.NewReviewRequestView."""

    fixtures = ['test_users']

    # TODO: Split this up into multiple unit tests, and do a better job of
    #       checking for expected results.
    def test_get(self):
        """Testing NewReviewRequestView.get"""
        with self.siteconfig_settings({'auth_require_sitewide_login': False},
                                      reload_settings=False):
            response = self.client.get('/r/new')
            self.assertEqual(response.status_code, 301)

            response = self.client.get('/r/new/')
            self.assertEqual(response.status_code, 302)

            self.client.login(username='grumpy', password='grumpy')

            response = self.client.get('/r/new/')
            self.assertEqual(response.status_code, 200)

    def test_read_only_mode_for_users(self):
        """Testing NewReviewRequestView when in read-only mode for regular
        users
        """
        self.siteconfig = SiteConfiguration.objects.get_current()
        settings = {
            'site_read_only': True,
        }

        with self.siteconfig_settings(settings):
            # Ensure user is redirected when trying to create new review
            # request.
            self.client.logout()
            self.client.login(username='doc', password='doc')

            resp = self.client.get('/r/new/')

            self.assertEqual(resp.status_code, 302)

    def test_read_only_mode_for_superusers(self):
        """Testing NewReviewRequestView when in read-only mode for superusers
        """
        self.siteconfig = SiteConfiguration.objects.get_current()
        settings = {
            'site_read_only': True,
        }

        with self.siteconfig_settings(settings):
            # Ensure admin can still access new while in read-only mode.
            self.client.logout()
            self.client.login(username='admin', password='admin')

            resp = self.client.get('/r/new/')

            self.assertEqual(resp.status_code, 200)

    def test_get_context_data_with_no_repos(self):
        """Testing NewReviewRequestView.get_context_data with no repositories
        """
        self.client.login(username='grumpy', password='grumpy')
        response = self.client.get('/r/new/')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['page_model_attrs'], {
            'localSitePrefix': '',
            'repositories': [],
        })

    @add_fixtures(['test_scmtools', 'test_site'])
    def test_get_context_data_with_repos(self):
        """Testing NewReviewRequestView.get_context_data with repositories"""
        self.client.login(username='grumpy', password='grumpy')
        user = User.objects.get(username='grumpy')

        # These will be shown in the repository list.
        repo1 = self.create_repository(
            name='Repository 1',
            tool_name='Git')
        repo2 = self.create_repository(
            name='Repository 2',
            tool_name='Subversion')
        repo3 = self.create_repository(
            name='Repository 3',
            tool_name='Perforce',
            public=False)

        repo3.users.add(user)

        # These won't be shown.
        self.create_repository(
            name='Repository 4',
            tool_name='Git',
            public=False)
        self.create_repository(
            name='Repository 5',
            tool_name='Git',
            with_local_site=True)

        response = self.client.get('/r/new/')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['page_model_attrs'], {
            'localSitePrefix': '',
            'repositories': [
                {
                    'filesOnly': False,
                    'id': repo1.pk,
                    'localSitePrefix': '',
                    'name': 'Repository 1',
                    'requiresBasedir': False,
                    'requiresChangeNumber': False,
                    'scmtoolName': 'Git',
                    'supportsPostCommit': False,
                },
                {
                    'filesOnly': False,
                    'id': repo2.pk,
                    'localSitePrefix': '',
                    'name': 'Repository 2',
                    'requiresBasedir': True,
                    'requiresChangeNumber': False,
                    'scmtoolName': 'Subversion',
                    'supportsPostCommit': True,
                },
                {
                    'filesOnly': False,
                    'id': repo3.pk,
                    'localSitePrefix': '',
                    'name': 'Repository 3',
                    'requiresBasedir': False,
                    'requiresChangeNumber': True,
                    'scmtoolName': 'Perforce',
                    'supportsPostCommit': False,
                },
            ],
        })

    @add_fixtures(['test_scmtools', 'test_site'])
    def test_get_context_data_with_repos_and_local_site(self):
        """Testing NewReviewRequestView.get_context_data with repositories
        and Local Site
        """
        user = User.objects.get(username='grumpy')
        self.get_local_site(self.local_site_name).users.add(user)

        self.client.login(username='grumpy', password='grumpy')

        # These will be shown in the repository list.
        repo1 = self.create_repository(
            name='Repository 1',
            tool_name='Git',
            with_local_site=True)
        repo2 = self.create_repository(
            name='Repository 2',
            tool_name='Subversion',
            with_local_site=True)
        repo3 = self.create_repository(
            name='Repository 3',
            tool_name='Perforce',
            public=False,
            with_local_site=True)

        repo3.users.add(user)

        # These won't be shown.
        self.create_repository(
            name='Repository 4',
            tool_name='Git',
            public=False,
            with_local_site=True)
        self.create_repository(
            name='Repository 5',
            tool_name='Git')

        local_site_prefix = 's/%s/' % self.local_site_name
        response = self.client.get('/%sr/new/' % local_site_prefix)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['page_model_attrs'], {
            'localSitePrefix': local_site_prefix,
            'repositories': [
                {
                    'filesOnly': False,
                    'id': repo1.pk,
                    'localSitePrefix': local_site_prefix,
                    'name': 'Repository 1',
                    'requiresBasedir': False,
                    'requiresChangeNumber': False,
                    'scmtoolName': 'Git',
                    'supportsPostCommit': False,
                },
                {
                    'filesOnly': False,
                    'id': repo2.pk,
                    'localSitePrefix': local_site_prefix,
                    'name': 'Repository 2',
                    'requiresBasedir': True,
                    'requiresChangeNumber': False,
                    'scmtoolName': 'Subversion',
                    'supportsPostCommit': True,
                },
                {
                    'filesOnly': False,
                    'id': repo3.pk,
                    'localSitePrefix': local_site_prefix,
                    'name': 'Repository 3',
                    'requiresBasedir': False,
                    'requiresChangeNumber': True,
                    'scmtoolName': 'Perforce',
                    'supportsPostCommit': False,
                },
            ],
        })



"""Unit tests for reviewboard.datagrids.columns.ShipItColumn.

Version Added:
    5.0.7
"""

from __future__ import annotations

from datetime import timedelta

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import ShipItColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class ShipItColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.ShipItColumn."""

    column = ShipItColumn()

    def test_render_data_with_none(self):
        """Testing ShipItColumn.render_data with 0 Ship It!'s, 0 issues"""
        review_request = self.create_review_request(publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_one_shipit(self):
        """Testing ShipItColumn.render_data with 1 Ship It!"""
        review_request = self.create_review_request(publish=True)
        review_request.shipit_count = 1
        review_request.last_review_activity_timestamp = (
            review_request.last_updated + timedelta(hours=1))

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="shipit-count-container"'
            '     title="1 Ship It!"'
            '     aria-label="1 Ship It!">'
            ' <span aria-hidden="true" class="shipit-count">'
            '  <span class="rb-icon rb-icon-datagrid-shipit"></span>'
            '  1'
            ' </span>'
            '</div>')

    def test_render_data_with_many_shipits(self):
        """Testing ShipItColumn.render_data with many Ship It!'s"""
        review_request = self.create_review_request(publish=True)
        review_request.shipit_count = 2
        review_request.last_review_activity_timestamp = (
            review_request.last_updated + timedelta(hours=1))

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="shipit-count-container"'
            '     title="2 Ship It\'s!"'
            '     aria-label="2 Ship It\'s!">'
            ' <span aria-hidden="true" class="shipit-count">'
            '  <span class="rb-icon rb-icon-datagrid-shipit"></span>'
            '  2'
            ' </span>'
            '</div>')

    def test_render_data_with_shipit_stale(self):
        """Testing ShipItColumn.render_data with Ship It!'s and stale"""
        review_request = self.create_review_request(publish=True)
        review_request.shipit_count = 1
        review_request.last_review_activity_timestamp = (
            review_request.last_updated - timedelta(hours=1))

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="shipit-count-container -is-stale"'
            '     title="1 Ship It! (New updates to review)"'
            '     aria-label="1 Ship It! (New updates to review)">'
            ' <span aria-hidden="true" class="shipit-count">'
            '  <span class="rb-icon rb-icon-datagrid-shipit"></span>'
            '  1'
            ' </span>'
            '</div>')

    def test_render_data_with_one_open_issue(self):
        """Testing ShipItColumn.render_data with one open issue"""
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 1
        review_request.shipit_count = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="issue-count-container"'
            '     title="1 issue opened"'
            '     aria-label="1 issue opened">'
            ' <span aria-hidden="true" class="issue-count">'
            '  <span class="rb-icon rb-icon-datagrid-open-issues"></span>'
            '  1'
            ' </span>'
            '</div>')

    def test_render_data_with_many_open_issues(self):
        """Testing ShipItColumn.render_data with many open issues"""
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 2
        review_request.shipit_count = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="issue-count-container"'
            '     title="2 issues opened"'
            '     aria-label="2 issues opened">'
            ' <span aria-hidden="true" class="issue-count">'
            '  <span class="rb-icon rb-icon-datagrid-open-issues"></span>'
            '  2'
            ' </span>'
            '</div>')

    def test_render_data_with_one_issue_pending_verification(self):
        """Testing ShipItColumn.render_data with one issue pending verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_verifying_count = 1
        review_request.shipit_count = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="issue-count-container"'
            '     title="1 issue requiring verification"'
            '     aria-label="1 issue requiring verification">'
            ' <span aria-hidden="true" class="issue-count">'
            '  <span class="rb-icon rb-icon-datagrid-issue-verifying"></span>'
            '  1'
            ' </span>'
            '</div>')

    def test_render_data_with_many_pending_verification(self):
        """Testing ShipItColumn.render_data with many issues pending
        verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_verifying_count = 2
        review_request.shipit_count = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="issue-count-container"'
            '     title="2 issues requiring verification"'
            '     aria-label="2 issues requiring verification">'
            ' <span aria-hidden="true" class="issue-count">'
            '  <span class="rb-icon rb-icon-datagrid-issue-verifying"></span>'
            '  2'
            ' </span>'
            '</div>')

    def test_render_data_with_one_issue_and_one_pending_verification(self):
        """Testing ShipItColumn.render_data with 1 open issue and pending
        verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 1
        review_request.issue_verifying_count = 1
        review_request.shipit_count = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="issue-count-container"'
            '     title="1 issue opened, 1 requiring verification"'
            '     aria-label="1 issue opened, 1 requiring verification">'
            ' <span aria-hidden="true" class="issue-count">'
            '  <span class="rb-icon rb-icon-datagrid-open-issues"></span>'
            '  1'
            ' </span>'
            ' <span aria-hidden="true" class="issue-verifying-count">'
            '  <span class="rb-icon rb-icon-datagrid-issue-verifying"></span>'
            '  1'
            ' </span>'
            '</div>')

    def test_render_data_with_many_issues_and_one_pending_verification(self):
        """Testing ShipItColumn.render_data with many open issues and 1
        pending verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 5
        review_request.issue_verifying_count = 1
        review_request.shipit_count = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="issue-count-container"'
            '     title="5 issues opened, 1 requiring verification"'
            '     aria-label="5 issues opened, 1 requiring verification">'
            ' <span aria-hidden="true" class="issue-count">'
            '  <span class="rb-icon rb-icon-datagrid-open-issues"></span>'
            '  5'
            ' </span>'
            ' <span aria-hidden="true" class="issue-verifying-count">'
            '  <span class="rb-icon rb-icon-datagrid-issue-verifying"></span>'
            '  1'
            ' </span>'
            '</div>')

    def test_render_data_with_many_issues_and_many_pending_verification(self):
        """Testing ShipItColumn.render_data with many open issues and many
        pending verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 5
        review_request.issue_verifying_count = 3
        review_request.shipit_count = 1

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<div class="issue-count-container"'
            '     title="5 issues opened, 3 requiring verification"'
            '     aria-label="5 issues opened, 3 requiring verification">'
            ' <span aria-hidden="true" class="issue-count">'
            '  <span class="rb-icon rb-icon-datagrid-open-issues"></span>'
            '  5'
            ' </span>'
            ' <span aria-hidden="true" class="issue-verifying-count">'
            '  <span class="rb-icon rb-icon-datagrid-issue-verifying"></span>'
            '  3'
            ' </span>'
            '</div>')

    def test_to_json_with_none(self) -> None:
        """Testing ShipItColumn.to_json with 0 Ship It!'s, 0 issues"""
        review_request = self.create_review_request(publish=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 0,
                'issue_verifying_count': 0,
                'shipit_count': 0,
                'shipits_stale': False,
            })

    def test_to_json_with_one_shipit(self) -> None:
        """Testing ShipItColumn.to_json with 1 Ship It!"""
        review_request = self.create_review_request(publish=True)
        review_request.shipit_count = 1
        review_request.last_review_activity_timestamp = (
            review_request.last_updated + timedelta(hours=1))

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 0,
                'issue_verifying_count': 0,
                'shipit_count': 1,
                'shipits_stale': False,
            })

    def test_to_json_with_many_shipits(self) -> None:
        """Testing ShipItColumn.to_json with many Ship It!'s"""
        review_request = self.create_review_request(publish=True)
        review_request.shipit_count = 2
        review_request.last_review_activity_timestamp = (
            review_request.last_updated + timedelta(hours=1))

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 0,
                'issue_verifying_count': 0,
                'shipit_count': 2,
                'shipits_stale': False,
            })

    def test_to_json_with_shipit_stale(self) -> None:
        """Testing ShipItColumn.to_json with Ship It!'s and stale"""
        review_request = self.create_review_request(publish=True)
        review_request.shipit_count = 1
        review_request.last_review_activity_timestamp = (
            review_request.last_updated - timedelta(hours=1))

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 0,
                'issue_verifying_count': 0,
                'shipit_count': 1,
                'shipits_stale': True,
            })

    def test_to_josn_with_one_open_issue(self) -> None:
        """Testing ShipItColumn.to_josn with one open issue"""
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 1
        review_request.shipit_count = 1

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 1,
                'issue_verifying_count': 0,
                'shipit_count': 1,
                'shipits_stale': False,
            })

    def test_to_json_with_many_open_issues(self) -> None:
        """Testing ShipItColumn.to_json with many open issues"""
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 2
        review_request.shipit_count = 1

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 2,
                'issue_verifying_count': 0,
                'shipit_count': 1,
                'shipits_stale': False,
            })

    def test_to_json_with_one_issue_pending_verification(self) -> None:
        """Testing ShipItColumn.to_json with one issue pending verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_verifying_count = 1
        review_request.shipit_count = 1

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 0,
                'issue_verifying_count': 1,
                'shipit_count': 1,
                'shipits_stale': False,
            })

    def test_to_json_with_many_pending_verification(self) -> None:
        """Testing ShipItColumn.to_json with many issues pending verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_verifying_count = 2
        review_request.shipit_count = 1

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 0,
                'issue_verifying_count': 2,
                'shipit_count': 1,
                'shipits_stale': False,
            })

    def test_to_json_with_one_issue_and_one_pending_verification(
        self,
    ) -> None:
        """Testing ShipItColumn.to_json with 1 open issue and pending
        verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 1
        review_request.issue_verifying_count = 1
        review_request.shipit_count = 1

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 1,
                'issue_verifying_count': 1,
                'shipit_count': 1,
                'shipits_stale': False,
            })

    def test_to_json_with_many_issues_and_one_pending_verification(
        self,
    ) -> None:
        """Testing ShipItColumn.to_json with many open issues and 1
        pending verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 5
        review_request.issue_verifying_count = 1
        review_request.shipit_count = 1

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 5,
                'issue_verifying_count': 1,
                'shipit_count': 1,
                'shipits_stale': False,
            })

    def test_to_json_with_many_issues_and_many_pending_verification(
        self,
    ) -> None:
        """Testing ShipItColumn.to_json with many open issues and many
        pending verification
        """
        review_request = self.create_review_request(publish=True)
        review_request.issue_open_count = 5
        review_request.issue_verifying_count = 3
        review_request.shipit_count = 1

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'issue_open_count': 5,
                'issue_verifying_count': 3,
                'shipit_count': 1,
                'shipits_stale': False,
            })



"""Unit tests for reviewboard.site.mixins.LocalSiteAwareModelFormMixin."""

from django import forms
from django.contrib.auth.models import User

from reviewboard.reviews.models import DefaultReviewer, Group
from reviewboard.site.mixins import LocalSiteAwareModelFormMixin
from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class LocalSiteAwareModelFormMixinTests(TestCase):
    """Unit tests for LocalSiteAwareModelFormMixin."""

    class MyForm(LocalSiteAwareModelFormMixin, forms.ModelForm):
        users = forms.ModelMultipleChoiceField(
            queryset=User.objects.filter(is_active=True))

        inactive_user = forms.ModelChoiceField(
            queryset=User.objects.filter(is_active=False))

        default_reviewer = forms.ModelChoiceField(
            queryset=DefaultReviewer.objects.all())

        class Meta:
            model = Group
            fields = '__all__'

    def setUp(self):
        super(LocalSiteAwareModelFormMixinTests, self).setUp()

        self.global_user = User.objects.create(username='global-user')
        self.site_user = User.objects.create(username='site-user')
        self.inactive_global_user = User.objects.create(
            username='inactive-global-user',
            is_active=False)
        self.inactive_site_user = User.objects.create(
            username='inactive-site-user',
            is_active=False)

        self.local_site = LocalSite.objects.create(name='site1')
        self.local_site.users.add(self.site_user, self.inactive_site_user)

        self.global_default_reviewer = DefaultReviewer.objects.create(
            name='global-default-reviewer',
            file_regex='.')
        self.site_default_reviewer = DefaultReviewer.objects.create(
            name='site-default-reviewer',
            file_regex='.',
            local_site=self.local_site)

    def test_without_localsite(self):
        """Testing LocalSiteAwareModelFormMixin without a LocalSite"""
        # Make sure the initial state and querysets are what we expect on init.
        form = self.MyForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.global_user, self.site_user])
        self.assertEqual(list(form.fields['inactive_user'].queryset),
                         [self.inactive_global_user, self.inactive_site_user])
        self.assertEqual(list(form.fields['default_reviewer'].queryset),
                         [self.global_default_reviewer,
                          self.site_default_reviewer])

        # Now test what happens when it's been fed data and validated.
        form = self.MyForm(data={
            'name': 'test-group',
            'display_name': 'Test Group',
            'users': [self.global_user.pk],
            'inactive_user': self.inactive_global_user.pk,
            'default_reviewer': self.global_default_reviewer.pk,
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.global_user, self.site_user])
        self.assertEqual(list(form.fields['inactive_user'].queryset),
                         [self.inactive_global_user, self.inactive_site_user])
        self.assertEqual(list(form.fields['default_reviewer'].queryset),
                         [self.global_default_reviewer,
                          self.site_default_reviewer])

        form.is_valid()
        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.global_user, self.site_user])
        self.assertEqual(list(form.fields['inactive_user'].queryset),
                         [self.inactive_global_user, self.inactive_site_user])
        self.assertEqual(list(form.fields['default_reviewer'].queryset),
                         [self.global_default_reviewer,
                          self.site_default_reviewer])

        new_group = form.save()
        self.assertEqual(list(new_group.users.all()), [self.global_user])
        self.assertIsNone(new_group.local_site_id)

    def test_without_localsite_and_edit_instance(self):
        """Testing LocalSiteAwareModelFormMixin without a LocalSite and
        editing an instance
        """
        group = self.create_review_group()

        form = self.MyForm(
            data={
                'name': 'new-group',
                'display_name': 'New Group',
                'users': [self.global_user.pk],
                'inactive_user': self.inactive_global_user.pk,
                'default_reviewer': self.global_default_reviewer.pk,
            },
            instance=group)
        self.assertTrue(form.is_valid())

        new_group = form.save()
        self.assertEqual(group.pk, new_group.pk)
        self.assertIsNone(new_group.local_site_id)

    def test_without_localsite_and_with_compatible_rel_values(self):
        """Testing LocalSiteAwareModelFormMixin without a LocalSite and
        compatible relation model values
        """
        # Note that Users are compatible even if on a Local Site, so long
        # as the form's model instance is not on a Local Site. However,
        # the DefaultReviewer is not compatible.
        form = self.MyForm(data={
            'name': 'new-group',
            'display_name': 'New Group',
            'users': [self.site_user.pk],
            'inactive_user': self.inactive_site_user.pk,
            'default_reviewer': self.global_default_reviewer.pk,
        })
        self.assertTrue(form.is_valid())

    def test_without_localsite_and_with_incompatible_rel_values(self):
        """Testing LocalSiteAwareModelFormMixin without a LocalSite and
        incompatible relation model values
        """
        form = self.MyForm(data={
            'name': 'new-group',
            'display_name': 'New Group',
            'users': [self.site_user.pk],
            'inactive_user': self.inactive_site_user.pk,
            'default_reviewer': self.site_default_reviewer.pk,
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'default_reviewer': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
            })

    def test_with_limited_localsite(self):
        """Testing LocalSiteAwareModelFormMixin limited to a LocalSite"""
        form = self.MyForm(limit_to_local_site=self.local_site)

        self.assertIs(form.limited_to_local_site, self.local_site)
        self.assertNotIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.site_user])
        self.assertEqual(list(form.fields['inactive_user'].queryset),
                         [self.inactive_site_user])
        self.assertEqual(list(form.fields['default_reviewer'].queryset),
                         [self.site_default_reviewer])

    def test_with_limited_localsite_and_changing_site(self):
        """Testing LocalSiteAwareModelFormMixin limited to a LocalSite and
        LocalSite in form data ignored
        """
        site2 = LocalSite.objects.create(name='test-site-2')

        form = self.MyForm(
            data={
                'name': 'new-group',
                'display_name': 'New Group',
                'users': [self.site_user.pk],
                'inactive_user': self.inactive_site_user.pk,
                'default_reviewer': self.site_default_reviewer.pk,
                'local_site': site2.pk,
            },
            limit_to_local_site=self.local_site)

        self.assertIs(form.limited_to_local_site, self.local_site)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['local_site'], self.local_site)

        group = form.save()
        self.assertEqual(group.local_site, self.local_site)

    def test_with_limited_localsite_and_compatible_instance(self):
        """Testing LocalSiteAwareModelFormMixin limited to a LocalSite and
        editing compatible instance
        """
        group = self.create_review_group(local_site=self.local_site)

        # This should just simply not raise an exception.
        self.MyForm(instance=group,
                    limit_to_local_site=self.local_site)

    def test_with_limited_localsite_and_incompatible_instance(self):
        """Testing LocalSiteAwareModelFormMixin limited to a LocalSite and
        editing incompatible instance
        """
        group = self.create_review_group()

        error_message = (
            'The provided instance is not associated with a LocalSite '
            'compatible with this form. Please contact support.'
        )

        # This should just simply not raise an exception.
        with self.assertRaisesMessage(ValueError, error_message):
            self.MyForm(instance=group,
                        limit_to_local_site=self.local_site)

    def test_with_limited_localsite_and_incompatible_rel_values(self):
        """Testing LocalSiteAwareModelFormMixin limited to a LocalSite and
        incompatible relation model values
        """
        form = self.MyForm(
            data={
                'name': 'new-group',
                'display_name': 'New Group',
                'users': [self.global_user.pk],
                'inactive_user': self.inactive_global_user.pk,
                'default_reviewer': self.global_default_reviewer.pk,
            },
            limit_to_local_site=self.local_site)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'default_reviewer': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
                'inactive_user': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
                'users': [
                    'Select a valid choice. 1 is not one of the available '
                    'choices.',
                ],
            })

    def test_with_localsite_in_data(self):
        """Testing LocalSiteAwareModelFormMixin with a LocalSite in form data
        """
        # Make sure the initial state and querysets are what we expect on init.
        form = self.MyForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.global_user, self.site_user])
        self.assertEqual(list(form.fields['inactive_user'].queryset),
                         [self.inactive_global_user, self.inactive_site_user])
        self.assertEqual(list(form.fields['default_reviewer'].queryset),
                         [self.global_default_reviewer,
                          self.site_default_reviewer])

        # Now test what happens when it's been fed data and validated.
        form = self.MyForm(data={
            'name': 'new-group',
            'display_name': 'New Group',
            'local_site': self.local_site.pk,
            'users': [self.site_user.pk],
            'inactive_user': self.inactive_site_user.pk,
            'default_reviewer': self.site_default_reviewer.pk,
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertTrue(form.is_valid())
        self.assertIn('local_site', form.fields)
        self.assertEqual(form.cleaned_data['local_site'], self.local_site)

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.global_user, self.site_user])
        self.assertEqual(list(form.fields['inactive_user'].queryset),
                         [self.inactive_global_user, self.inactive_site_user])
        self.assertEqual(list(form.fields['default_reviewer'].queryset),
                         [self.global_default_reviewer,
                          self.site_default_reviewer])

        group = form.save()
        self.assertEqual(group.local_site, self.local_site)
        self.assertEqual(list(group.users.all()), [self.site_user])

    def test_with_localsite_in_data_and_edit_instance(self):
        """Testing LocalSiteAwareModelFormMixin with a LocalSite in form data
        and editing instance
        """
        group = self.create_review_group()

        form = self.MyForm(
            data={
                'name': 'new-group',
                'display_name': 'New Group',
                'local_site': self.local_site.pk,
                'users': [self.site_user.pk],
                'inactive_user': self.inactive_site_user.pk,
                'default_reviewer': self.site_default_reviewer.pk,
            },
            instance=group)
        self.assertTrue(form.is_valid())

        new_group = form.save()
        self.assertEqual(new_group.pk, group.pk)
        self.assertEqual(new_group.local_site, self.local_site)
        self.assertEqual(list(new_group.users.all()), [self.site_user])

    def test_with_localsite_in_data_and_incompatible_rel_values(self):
        """Testing LocalSiteAwareModelFormMixin with a LocalSite in form data
        and incompatible relation model values
        """
        form = self.MyForm(data={
            'name': 'new-group',
            'display_name': 'New Group',
            'local_site': self.local_site.pk,
            'users': [self.global_user.pk],
            'inactive_user': self.inactive_global_user.pk,
            'default_reviewer': self.global_default_reviewer.pk,
        })

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'default_reviewer': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
                'inactive_user': [
                    'Select a valid choice. That choice is not one of the '
                    'available choices.',
                ],
                'users': [
                    'Select a valid choice. 1 is not one of the available '
                    'choices.',
                ],
            })

    def test_with_localsite_in_data_with_bad_value(self):
        """Testing LocalSiteAwareModelFormMixin with a LocalSite in form data
        and ID is a non-integer
        """
        # This should just not crash.
        form = self.MyForm(data={
            'name': 'new-group',
            'display_name': 'New Group',
            'local_site': 'abc',
        })

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors['local_site'],
            [
                'Select a valid choice. That choice is not one of the '
                'available choices.',
            ])



"""Diff chunk generator implementations."""

from __future__ import annotations

import fnmatch
import hashlib
import logging
import re
from itertools import zip_longest
from typing import Optional, Sequence, TYPE_CHECKING, Union

import pygments.util
from django.utils.encoding import force_str
from django.utils.html import escape
from django.utils.translation import get_language, gettext as _
from djblets.log import log_timed
from djblets.cache.backend import cache_memoize
from housekeeping.functions import deprecate_non_keyword_only_args
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import (find_lexer_class,
                             guess_lexer_for_filename)

from reviewboard.codesafety import code_safety_checker_registry
from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.diffviewer.differ import DiffCompatVersion, get_differ
from reviewboard.diffviewer.diffutils import (get_filediff_encodings,
                                              get_line_changed_regions,
                                              get_original_file,
                                              get_original_and_patched_files,
                                              get_patched_file,
                                              get_sha256,
                                              convert_to_unicode,
                                              split_line_endings)
from reviewboard.diffviewer.opcode_generator import get_diff_opcode_generator
from reviewboard.diffviewer.settings import DiffSettings

if TYPE_CHECKING:
    from django.http import HttpRequest

    from reviewboard.diffviewer.models.filediff import FileDiff


logger = logging.getLogger(__name__)


class NoWrapperHtmlFormatter(HtmlFormatter):
    """An HTML Formatter for Pygments that doesn't wrap items in a div."""
    def __init__(self, *args, **kwargs):
        super(NoWrapperHtmlFormatter, self).__init__(*args, **kwargs)

    def _wrap_div(self, inner):
        """Removes the div wrapper from formatted code.

        This is called by the formatter to wrap the contents of inner.
        Inner is a list of tuples containing formatted code. If the first item
        in the tuple is zero, then it's the div wrapper, so we should ignore
        it.
        """
        for tup in inner:
            if tup[0]:
                yield tup


class RawDiffChunkGenerator(object):
    """A generator for chunks for a diff that can be used for rendering.

    Each chunk represents an insert, delete, replace, or equal section. It
    contains all the data needed to render the portion of the diff.

    This is general-purpose and meant to operate on strings each consisting of
    at least one line of text, or lists of lines of text.

    If the caller passes lists of lines instead of strings, then the
    caller will also be responsible for applying any syntax highlighting and
    dealing with newline differences.

    Chunk generator instances must be recreated for each new file being
    parsed.

    Version Changed:
        5.0:
        Added :py:attr:`all_code_safety_results`.

    Attributes:
        all_code_safety_results (dict):
            Code safety warnings were found while processing the diff.

            This is in the form of::

                {
                    '<checker_id>': {
                        'warnings': {'<result_id>', ...},
                        'errors': {'<result_id>', ...},
                    },
                    ...
                }

            All keys are optional.

            Version Added:
                5.0
    """

    NEWLINES_RE = re.compile(r'\r?\n')

    # The maximum size a line can be before we start shutting off styling.
    STYLED_MAX_LINE_LEN = 1000
    STYLED_MAX_LIMIT_BYTES = 200000  # 200KB

    # A list of filename extensions that won't be styled.
    STYLED_EXT_BLACKLIST = (
        '.txt',  # ResourceLexer is used as a default.
    )

    #: The default width for a tabstop.
    TAB_SIZE = DiffSettings.DEFAULT_TAB_SIZE

    ######################
    # Instance variables #
    ######################

    #: Settings used for the generation of the diff.
    #:
    #: Version Added:
    #:     5.0.2
    #:
    #: Type:
    #:     reviewboard.diffviewer.settings.DiffSettings
    diff_settings: DiffSettings

    def __init__(
        self,
        old: Optional[Union[bytes, Sequence[bytes]]],
        new: Optional[Union[bytes, Sequence[bytes]]],
        orig_filename: str,
        modified_filename: str,
        encoding_list: Optional[Sequence[str]] = None,
        diff_compat: int = DiffCompatVersion.DEFAULT,
        *,
        diff_settings: DiffSettings,
    ) -> None:
        """Initialize the chunk generator.

        Version Changed:
            6.0:
            * Removed ``enable_syntax_highlighting``.
            * Made ``diff_settings`` mandatory.

        Version Changed:
            5.0.2:
            * Added ``diff_settings``, which will be required starting in
              Review Board 6.
            * Deprecated ``enable_syntax_highlighting`` in favor of
              ``diff_settings``.

        Args:
            old (bytes or list of bytes):
                The old data being modified.

            new (bytes or list of bytes):
                The new data.

            orig_filename (unicode):
                The filename corresponding to the old data.

            modified_filename (unicode):
                The filename corresponding to the new data.

            encoding_list (list of unicode, optional):
                A list of encodings to try for the ``old`` and ``new`` data,
                when converting to Unicode. If not specified, this defaults
                to ``iso-8859-15``.

            diff_compat (int, optional):
                A specific diff compatibility version to use for any diffing
                logic.

            diff_settings (reviewboard.diffviewer.settings.DiffSettings):
                The settings used to control the display of diffs.

                Version Added:
                    5.0.2
        """
        # Check that the data coming in is in the formats we accept.
        for param, param_name in ((old, 'old'), (new, 'new')):
            if param is not None:
                if isinstance(param, list):
                    if param and not isinstance(param[0], bytes):
                        raise TypeError(
                            _('%s expects None, list of bytes, or bytes '
                              'value for "%s", not list of %s')
                            % (type(self).__name__, param_name,
                               type(param[0])))
                elif not isinstance(param, bytes):
                    raise TypeError(
                        _('%s expects None, list of bytes, or bytes value '
                          'for "%s", not %s')
                        % (type(self).__name__, param_name, type(param)))

        if not isinstance(orig_filename, str):
            raise TypeError(
                _('%s expects a Unicode value for "orig_filename"')
                % type(self).__name__)

        if not isinstance(modified_filename, str):
            raise TypeError(
                _('%s expects a Unicode value for "modified_filename"')
                % type(self).__name__)

        assert diff_settings.tab_size
        self.diff_settings = diff_settings

        self.old = old
        self.new = new
        self.orig_filename = orig_filename
        self.modified_filename = modified_filename
        self.diff_settings = diff_settings
        self.enable_syntax_highlighting = diff_settings.syntax_highlighting
        self.encoding_list = encoding_list or ['iso-8859-15']
        self.diff_compat = diff_compat
        self.differ = None

        self.all_code_safety_results = {}

        # Chunk processing state.
        self._last_header = [None, None]
        self._last_header_index = [0, 0]
        self._chunk_index = 0

    def get_opcode_generator(self):
        """Return the DiffOpcodeGenerator used to generate diff opcodes."""
        return get_diff_opcode_generator(self.differ)

    def get_chunks(self, cache_key=None):
        """Return the chunks for the given diff information.

        If a cache key is provided and there are chunks already computed in the
        cache, they will be yielded. Otherwise, new chunks will be generated,
        stored in cache (given a cache key), and yielded.
        """
        if cache_key:
            chunks = cache_memoize(cache_key,
                                   lambda: list(self.get_chunks_uncached()),
                                   large_data=True)
        else:
            chunks = self.get_chunks_uncached()

        for chunk in chunks:
            yield chunk

    def get_chunks_uncached(self):
        """Yield the list of chunks, bypassing the cache."""
        for chunk in self.generate_chunks(self.old, self.new):
            yield chunk

    def generate_chunks(self, old, new, old_encoding_list=None,
                        new_encoding_list=None):
        """Generate chunks for the difference between two strings.

        The strings will be normalized, ensuring they're of the proper
        encoding and ensuring they have consistent newlines. They're then
        syntax-highlighted (if requested).

        Once the strings are ready, chunks are built from the strings and
        yielded to the caller. Each chunk represents information on an
        equal, inserted, deleted, or replaced set of lines.

        The number of lines of each chunk type are stored in the
        :py:attr:`counts` dictionary, which can then be accessed after
        yielding all chunks.

        Args:
            old (bytes or list of bytes):
                The old data being modified.

            new (bytes or list of bytes):
                The new data.

            old_encoding_list (list of unicode, optional):
                An optional list of encodings that ``old`` may be encoded in.
                If not provided, :py:attr:`encoding_list` is used.

            new_encoding_list (list of unicode, optional):
                An optional list of encodings that ``new`` may be encoded in.
                If not provided, :py:attr:`encoding_list` is used.

        Yields:
            dict:
            A rendered chunk containing the following keys:

            ``index`` (int)
                The 0-based index of the chunk.

            ``lines`` (list of unicode):
                The rendered list of lines.

            ``numlines`` (int):
                The number of lines in the chunk.

            ``change`` (unicode):
                The type of change (``delete``, ``equal``, ``insert`` or
                ``replace``).

            ``collapsable`` (bool):
                Whether the chunk can be collapsed.

            ``meta`` (dict):
                Metadata on the chunk.
        """
        is_lists = isinstance(old, list)
        assert is_lists == isinstance(new, list)

        if old_encoding_list is None:
            old_encoding_list = self.encoding_list

        if new_encoding_list is None:
            new_encoding_list = self.encoding_list

        if is_lists:
            if self.encoding_list:
                old = self.normalize_source_list(old, old_encoding_list)
                new = self.normalize_source_list(new, new_encoding_list)

            a = old
            b = new
        else:
            old, a = self.normalize_source_string(old, old_encoding_list)
            new, b = self.normalize_source_string(new, new_encoding_list)

        a_num_lines = len(a)
        b_num_lines = len(b)

        if is_lists:
            markup_a = a
            markup_b = b
        else:
            markup_a = None
            markup_b = None

            if self._get_enable_syntax_highlighting(old, new, a, b):
                # TODO: Try to figure out the right lexer for these files
                #       once instead of twice.
                markup_a = self._apply_pygments(
                    old or '',
                    self.normalize_path_for_display(self.orig_filename))
                markup_b = self._apply_pygments(
                    new or '',
                    self.normalize_path_for_display(self.modified_filename))

            if not markup_a:
                markup_a = self.NEWLINES_RE.split(escape(old))

            if not markup_b:
                markup_b = self.NEWLINES_RE.split(escape(new))

        ignore_space = True

        for pattern in self.diff_settings.include_space_patterns:
            if fnmatch.fnmatch(self.orig_filename, pattern):
                ignore_space = False
                break

        self.differ = get_differ(a, b, ignore_space=ignore_space,
                                 compat_version=self.diff_compat)
        self.differ.add_interesting_lines_for_headers(self.orig_filename)

        context_num_lines = self.diff_settings.context_num_lines
        collapse_threshold = 2 * context_num_lines + 3

        line_num = 1
        opcodes_generator = self.get_opcode_generator()

        counts = {
            'equal': 0,
            'replace': 0,
            'insert': 0,
            'delete': 0,
        }

        for tag, i1, i2, j1, j2, meta in opcodes_generator:
            old_lines = markup_a[i1:i2]
            new_lines = markup_b[j1:j2]
            num_lines = max(len(old_lines), len(new_lines))

            lines = [
                self._diff_line(tag, meta, *diff_args)
                for diff_args in zip_longest(
                    range(line_num, line_num + num_lines),
                    range(i1 + 1, i2 + 1),
                    range(j1 + 1, j2 + 1),
                    a[i1:i2],
                    b[j1:j2],
                    old_lines,
                    new_lines)
            ]

            counts[tag] += num_lines

            if tag == 'equal' and num_lines > collapse_threshold:
                last_range_start = num_lines - context_num_lines

                if line_num == 1:
                    yield self._new_chunk(lines, 0, last_range_start, True)
                    yield self._new_chunk(lines, last_range_start, num_lines)
                else:
                    yield self._new_chunk(lines, 0, context_num_lines)

                    if i2 == a_num_lines and j2 == b_num_lines:
                        yield self._new_chunk(lines, context_num_lines,
                                              num_lines, True)
                    else:
                        yield self._new_chunk(lines, context_num_lines,
                                              last_range_start, True)
                        yield self._new_chunk(lines, last_range_start,
                                              num_lines)
            else:
                yield self._new_chunk(lines, 0, num_lines, False, tag, meta)

            line_num += num_lines

        self.counts = counts

    def check_line_code_safety(self, orig_line, modified_line,
                               extra_state={}, **kwargs):
        """Check the safety of a line of code.

        This will run the original and modified line through all registered
        code safety checkers. If any checker produces warnings or errors,
        those will be associated with the line.

        Version Added:
            5.0

        Args:
            orig_line (str):
                The original line to check.

            modified_line (str):
                The modiifed line to check.

            extra_state (dict, optional):
                Extra state to pass to the checker for the original or
                modified line content item. Used by subclasses to produce
                additional information that may be useful for some code safety
                checkers.

            **kwargs (dict, optional):
                Unused keyword arguments, for future expansion.

        Returns:
            list of tuple:
            A list of code safety results containing warnings or errors. Each
            item is a tuple containing:

            1. The registered checker ID.
            2. A dictionary with ``errors`` and/or ``warnings`` keys.
        """
        # Check for any unsafe/suspicious content on this line by passing
        # the raw source through any registered code safety checkers.
        results = []
        to_check = []

        if orig_line:
            to_check.append(dict({
                'path': self.orig_filename,
                'lines': [orig_line],
            }, **extra_state))

        if modified_line:
            to_check.append(dict({
                'path': self.modified_filename,
                'lines': [modified_line],
            }, **extra_state))

        if to_check:
            code_safety_configs = self.diff_settings.code_safety_configs

            # We have code to check. Let's go through each code checker
            # and see if we have any warnings or errors to display.
            for checker in code_safety_checker_registry:
                checker_config = code_safety_configs.get(checker.checker_id,
                                                         {})
                checker_results = checker.check_content(content_items=to_check,
                                                        **checker_config)

                if checker_results:
                    # Normalize the code checker results. We're going to
                    # extract only "errors" and "warnings" keys and set them
                    # only if they have a truthy value (present in the
                    # dictionary, non-None, and not an empty list).
                    norm_checker_results = {}

                    for key in ('errors', 'warnings'):
                        checker_result_ids = checker_results.get(key)

                        if checker_result_ids:
                            norm_checker_results[key] = checker_result_ids

                    if norm_checker_results:
                        # We have normalized checker results, so add them to
                        # the final list of results as documented in "Returns".
                        results.append((checker.checker_id,
                                        norm_checker_results))

        return results

    def normalize_source_string(self, s, encoding_list, **kwargs):
        """Normalize a source string of text to use for the diff.

        This will normalize the encoding of the string and the newlines,
        returning a tuple containing the normalized string and a list of
        lines split from the source.

        Both the original and modified strings used for the diff will be
        normalized independently.

        This is only used if the caller passes a string instead of a list for
        the original or new values.

        Subclasses can override this to provide custom behavior.

        Args:
            s (bytes):
                The string to normalize.

            encoding_list (list of unicode):
                The list of encodings to try when converting the string to
                Unicode.

            **kwargs (dict):
                Additional keyword arguments, for future expansion.

        Returns:
            tuple:
            A tuple containing:

            1. The full normalized string
            2. The list of lines from the string

        Raises:
            UnicodeDecodeError:
                The string could not be converted to Unicode.
        """
        s = convert_to_unicode(s, encoding_list)[1]

        # Normalize the input so that if there isn't a trailing newline, we
        # add it.
        if s and not s.endswith('\n'):
            s += '\n'

        lines = self.NEWLINES_RE.split(s or '')

        # Remove the trailing newline, now that we've split this. This will
        # prevent a duplicate line number at the end of the diff.
        del lines[-1]

        return s, lines

    def normalize_source_list(self, lines, encoding_list, **kwargs):
        """Normalize a list of source lines to use for the diff.

        This will normalize the encoding of the lines.

        Both the original and modified lists of lines used for the diff will be
        normalized independently.

        This is only used if the caller passes a list instead of a string for
        the original or new values.

        Subclasses can override this to provide custom behavior.

        Args:
            lines (list of bytes):
                The list of lines to normalize.

            encoding_list (list of unicode):
                The list of encodings to try when converting the lines to
                Unicode.

            **kwargs (dict):
                Additional keyword arguments, for future expansion.

        Returns:
            list of unicode:
            The resulting list of normalized lines.

        Raises:
            UnicodeDecodeError:
                One or more lines could not be converted to Unicode.
        """
        if encoding_list:
            lines = [
                convert_to_unicode(s, encoding_list)[1]
                for s in lines
            ]

        return lines

    def normalize_path_for_display(self, filename):
        """Normalize a file path for display to the user.

        By default, this returns the filename as-is. Subclasses can override
        the behavior to return a variant of the filename.

        Args:
            filename (unicode):
                The filename to normalize.

        Returns:
            unicode:
            The normalized filename.
        """
        return filename

    def get_line_changed_regions(self, old_line_num, old_line,
                                 new_line_num, new_line):
        """Return information on changes between two lines.

        This returns a tuple containing a list of tuples of ranges in the
        old line, and a list of tuples of ranges in the new line, that
        should be highlighted.

        This defaults to simply wrapping get_line_changed_regions() from
        diffutils. Subclasses can override to provide custom behavior.
        """
        return get_line_changed_regions(old_line, new_line)

    def _get_enable_syntax_highlighting(self, old, new, a, b):
        """Returns whether or not we'll be enabling syntax highlighting.

        This is based first on the value received when constructing the
        generator, and then based on heuristics to determine if it's fast
        enough to render with syntax highlighting on.

        The heuristics take into account the size of the files in bytes and
        the number of lines.
        """
        if not self.enable_syntax_highlighting:
            return False

        threshold = self.diff_settings.syntax_highlighting_threshold

        if threshold and (len(a) > threshold or len(b) > threshold):
            return False

        # Very long files, especially XML files, can take a long time to
        # highlight. For files over a certain size, don't highlight them.
        if (len(old) > self.STYLED_MAX_LIMIT_BYTES or
                len(new) > self.STYLED_MAX_LIMIT_BYTES):
            return False

        # Don't style the file if we have any *really* long lines.
        # It's likely a minified file or data or something that doesn't
        # need styling, and it will just grind Review Board to a halt.
        for lines in (a, b):
            for line in lines:
                if len(line) > self.STYLED_MAX_LINE_LEN:
                    return False

        return True

    def _diff_line(self, tag, meta, v_line_num, old_line_num, new_line_num,
                   old_line, new_line, old_markup, new_markup):
        """Creates a single line in the diff viewer.

        Information on the line will be returned, and later will be used
        for rendering the line. The line represents a single row of a
        side-by-side diff. It contains a row number, real line numbers,
        region information, syntax-highlighted HTML for the text,
        and other metadata.
        """
        if (tag == 'replace' and
            old_line and new_line and
            len(old_line) <= self.STYLED_MAX_LINE_LEN and
            len(new_line) <= self.STYLED_MAX_LINE_LEN and
            old_line != new_line):
            # Generate information on the regions that changed between the
            # two lines.
            old_region, new_region = \
                self.get_line_changed_regions(old_line_num, old_line,
                                              new_line_num, new_line)
        else:
            old_region = new_region = []

        old_markup = old_markup or ''
        new_markup = new_markup or ''

        line_pair = (old_line_num, new_line_num)

        indentation_changes = meta.get('indentation_changes', {})

        if line_pair[0] is not None and line_pair[1] is not None:
            indentation_change = indentation_changes.get('%d-%d' % line_pair)

            # We check the ranges against (0, 0) for compatibility with a bug
            # present in Review Board 4.0.6 and older, where bad indentation
            # calculation logic could incorrectly determine that two
            # "filtered-equal" lines in interdiffs had a 0-length indentation
            # change. This broke our serialization logic.
            #
            # Review Board 4.0.7 and higher address this problem, but we could
            # be showing something that's still in cache. Note however that
            # the "indentation" lines will still be broken up into their own
            # chunks in the diff viewer, but at least they'll render.
            if indentation_change and indentation_change[1:] != (0, 0):
                old_markup, new_markup = self._highlight_indentation(
                    old_markup, new_markup, *indentation_change)

        result = [
            v_line_num,
            old_line_num or '', old_markup, old_region,
            new_line_num or '', new_markup, new_region,
            line_pair in meta['whitespace_lines']
        ]

        # NOTE: Prior to Review Board 5, this only contained moved info
        #       ("to"/"from" keys), and was not used for general line-level
        #       metadata. To avoid changing the structure of the line format
        #       too much (given that this can be consumed by third-parties),
        #       we have updated this to be a general-purpose metadata storage.
        line_meta = {}

        # Record all the moved line numbers, carefully making note of the
        # start of each range. Ranges start when the previous line number is
        # either not in a move range or does not immediately precede this
        # line.
        for direction, moved_line_num in (('to', old_line_num),
                                          ('from', new_line_num)):
            moved_meta = meta.get(f'moved-{direction}', {})
            direction_move_info = self._get_move_info(moved_line_num,
                                                      moved_meta)

            if direction_move_info is not None:
                line_meta[direction] = direction_move_info

        # Check for any unsafe/suspicious content on this line by passing
        # the raw source through any registered code safety checkers.
        code_safety_results = self.check_line_code_safety(
            orig_line=old_line,
            modified_line=new_line)

        if code_safety_results:
            # Store the code safety information for this line and in the
            # diff-wide all_code_safety_results.
            line_meta['code_safety'] = code_safety_results
            all_code_safety_results = self.all_code_safety_results

            for checker_id, checker_results in code_safety_results:
                all_code_checker_results = \
                    all_code_safety_results.setdefault(checker_id, {})

                for key, checker_result_ids in checker_results.items():
                    all_code_checker_results.setdefault(key, set()).update(
                        checker_result_ids)

        # Only include the meta information for the line if it has content.
        # Otherwise, save the space in cache.
        if line_meta:
            result.append(line_meta)

        return result

    def _get_move_info(self, line_num, moved_meta):
        """Return information for a moved line.

        This will return a tuple containing the line number on the other end
        of the move for a line, and whether this is the beginning of a move
        range.

        Args:
            line_num (int):
                The line number that was part of a move.

            moved_meta (dict):
                Information on the move.

        Returns:
            tuple:
            A tuple of ``(other_line, is_first_in_range)``.
        """
        if not line_num or line_num not in moved_meta:
            return None

        other_line_num = moved_meta[line_num]

        return (
            other_line_num,
            (line_num - 1 not in moved_meta or
             other_line_num != moved_meta[line_num - 1] + 1)
        )

    def _highlight_indentation(self, old_markup, new_markup, is_indent,
                               raw_indent_len, norm_indent_len_diff):
        """Highlights indentation in an HTML-formatted line.

        This will wrap the indentation in <span> tags, and format it in
        a way that makes it clear how many spaces or tabs were used.
        """
        if is_indent:
            new_markup = self._wrap_indentation_chars(
                'indent',
                new_markup,
                raw_indent_len,
                norm_indent_len_diff,
                self._serialize_indentation)
        else:
            old_markup = self._wrap_indentation_chars(
                'unindent',
                old_markup,
                raw_indent_len,
                norm_indent_len_diff,
                self._serialize_unindentation)

        return old_markup, new_markup

    def _wrap_indentation_chars(self, class_name, markup, raw_indent_len,
                                norm_indent_len_diff, serializer):
        """Wraps characters in a string with indentation markers.

        This will insert the indentation markers and its wrapper in the
        markup string. It's careful not to interfere with any tags that
        may be used to highlight that line.
        """
        start_pos = 0

        # There may be a tag wrapping this whitespace. If so, we need to
        # find where the actual whitespace chars begin.
        while markup[start_pos] == '<':
            end_tag_pos = markup.find('>', start_pos + 1)

            # We'll only reach this if some corrupted HTML was generated.
            # We want to know about that.
            assert end_tag_pos != -1

            start_pos = end_tag_pos + 1

        end_pos = start_pos + raw_indent_len

        indentation = markup[start_pos:end_pos]

        if indentation.strip() != '':
            # There may be other things in here we didn't expect. It's not
            # a straight sequence of characters. Give up on highlighting it.
            return markup

        serialized, remainder = serializer(indentation, norm_indent_len_diff)

        return '%s<span class="%s">%s</span>%s' % (
            markup[:start_pos],
            class_name,
            serialized,
            remainder + markup[end_pos:])

    def _serialize_indentation(self, chars, norm_indent_len_diff):
        """Serializes an indentation string into an HTML representation.

        This will show every space as ``>``, and every tab as ``------>|``.
        In the case of tabs, we display as much of it as possible (anchoring
        to the right-hand side) given the space we have within the tab
        boundary.

        Args:
            chars (unicode):
                The indentation characters to serialize.

            norm_indent_len_diff (int):
                The difference in indentation between the old and new lines.

        Returns:
            tuple:
            A 2-tuple containing:

            1. The serialized indentation string.
            2. The remaining indentation characters not serialized.
        """
        assert chars

        s = ''
        i = 0
        j = 0

        tab_size = self.diff_settings.tab_size
        assert tab_size

        for j, c in enumerate(chars):
            if c == ' ':
                s += '&gt;'
                i += 1
            elif c == '\t':
                # Build "------>|" with the room we have available.
                in_tab_pos = i % tab_size

                if in_tab_pos < tab_size - 1:
                    if in_tab_pos < tab_size - 2:
                        num_dashes = (tab_size - 2 - in_tab_pos)
                        s += '&mdash;' * num_dashes
                        i += num_dashes

                    s += '&gt;'
                    i += 1

                s += '|'
                i += 1

            if i >= norm_indent_len_diff:
                break

        return s, chars[j + 1:]

    def _serialize_unindentation(self, chars, norm_indent_len_diff):
        """Serialize an unindentation string into an HTML representation.

        This will show every space as ``<``, and every tab as ``|<------``.
        In the case of tabs, we display as much of it as possible (anchoring
        to the left-hand side) given the space we have within the tab
        boundary.

        Args:
            chars (unicode):
                The unindentation characters to serialize.

            norm_indent_len_diff (int):
                The difference in indentation between the old and new lines.

        Returns:
            tuple:
            A 2-tuple containing:

            1. The serialized unindentation string.
            2. The remaining unindentation characters not serialized.
        """
        assert chars

        s = ''
        i = 0
        j = 0

        tab_size = self.diff_settings.tab_size
        assert tab_size

        for j, c in enumerate(chars):
            if c == ' ':
                s += '&lt;'
                i += 1
            elif c == '\t':
                # Build "|<------" with the room we have available.
                in_tab_pos = i % tab_size

                s += '|'
                i += 1

                if in_tab_pos < tab_size - 1:
                    s += '&lt;'
                    i += 1

                    if in_tab_pos < tab_size - 2:
                        num_dashes = (tab_size - 2 - in_tab_pos)
                        s += '&mdash;' * num_dashes
                        i += num_dashes

            if i >= norm_indent_len_diff:
                break

        return s, chars[j + 1:]

    def _new_chunk(self, all_lines, start, end, collapsable=False,
                   tag='equal', meta=None):
        """Creates a chunk.

        A chunk represents an insert, delete, or equal region. The chunk
        contains a bunch of metadata for things like whether or not it's
        collapsable and any header information.

        This is what ends up being returned to the caller of this class.
        """
        if not meta:
            meta = {}

        left_headers = list(self._get_interesting_headers(
            all_lines, start, end - 1, False))
        right_headers = list(self._get_interesting_headers(
            all_lines, start, end - 1, True))

        meta['left_headers'] = left_headers
        meta['right_headers'] = right_headers

        lines = all_lines[start:end]
        num_lines = len(lines)

        compute_chunk_last_header(lines, num_lines, meta, self._last_header)

        if (collapsable and end < len(all_lines) and
                (self._last_header[0] or self._last_header[1])):
            meta['headers'] = list(self._last_header)

        chunk = {
            'index': self._chunk_index,
            'lines': lines,
            'numlines': num_lines,
            'change': tag,
            'collapsable': collapsable,
            'meta': meta,
        }

        self._chunk_index += 1

        return chunk

    def _get_interesting_headers(self, lines, start, end, is_modified_file):
        """Returns all headers for a region of a diff.

        This scans for all headers that fall within the specified range
        of the specified lines on both the original and modified files.
        """
        possible_functions = \
            self.differ.get_interesting_lines('header', is_modified_file)

        if not possible_functions:
            return

        try:
            if is_modified_file:
                last_index = self._last_header_index[1]
                i1 = lines[start][4]
                i2 = lines[end - 1][4]
            else:
                last_index = self._last_header_index[0]
                i1 = lines[start][1]
                i2 = lines[end - 1][1]
        except IndexError:
            return

        for i in range(last_index, len(possible_functions)):
            linenum, line = possible_functions[i]
            linenum += 1

            if i2 != '' and linenum > i2:
                break
            elif i1 != '' and linenum >= i1:
                last_index = i
                yield linenum, line

        if is_modified_file:
            self._last_header_index[1] = last_index
        else:
            self._last_header_index[0] = last_index

    def _apply_pygments(self, data, filename):
        """Apply Pygments syntax-highlighting to a file's contents.

        This will only apply syntax highlighting if a lexer is available and
        the file extension is not blacklisted.

        Args:
            data (unicode):
                The data to syntax highlight.

            filename (unicode):
                The name of the file. This is used to help determine a
                suitable lexer.

        Returns:
            list of unicode:
            A list of lines, all syntax-highlighted, if a lexer is found.
            If no lexer is available, this will return ``None``.
        """
        if filename.endswith(self.STYLED_EXT_BLACKLIST):
            return None

        custom_pygments_lexers = self.diff_settings.custom_pygments_lexers
        lexer = None

        for ext, lexer_name in custom_pygments_lexers.items():
            if ext and filename.endswith(ext):
                lexer_class = find_lexer_class(lexer_name)

                if lexer_class:
                    lexer = lexer_class(stripnl=False,
                                        encoding='utf-8')
                    break
                else:
                    logger.error(
                        'Pygments lexer "%s" for "%s" files in '
                        'Diff Viewer Settings was not found.',
                        lexer_name, ext)
        else:
            try:
                lexer = guess_lexer_for_filename(filename,
                                                 data,
                                                 stripnl=False,
                                                 encoding='utf-8')
            except pygments.util.ClassNotFound:
                return None

        lexer.add_filter('codetagify')

        return split_line_endings(
            highlight(data, lexer, NoWrapperHtmlFormatter()))


class DiffChunkGenerator(RawDiffChunkGenerator):
    """A generator for chunks for a FileDiff that can be used for rendering.

    Each chunk represents an insert, delete, replace, or equal section. It
    contains all the data needed to render the portion of the diff.

    There are three ways this can operate, based on provided parameters.

    1) filediff, no interfilediff -
       Returns chunks for a single filediff. This is the usual way
       people look at diffs in the diff viewer.

       In this mode, we get the original file based on the filediff
       and then patch it to get the resulting file.

       This is also used for interdiffs where the source revision
       has no equivalent modified file but the interdiff revision
       does. It's no different than a standard diff.

    2) filediff, interfilediff -
       Returns chunks showing the changes between a source filediff
       and the interdiff.

       This is the typical mode used when showing the changes
       between two diffs. It requires that the file is included in
       both revisions of a diffset.

    3) filediff, no interfilediff, force_interdiff -
       Returns chunks showing the changes between a source
       diff and an unmodified version of the diff.

       This is used when the source revision in the diffset contains
       modifications to a file which have then been reverted in the
       interdiff revision. We don't actually have an interfilediff
       in this case, so we have to indicate that we are indeed in
       interdiff mode so that we can special-case this and not
       grab a patched file for the interdiff version.
    """

    ######################
    # Instance variables #
    ######################

    #: The original lines for the start revision of the interdiff.
    #:
    #: Version Added:
    #:     8.0
    _interdiff_filediff_orig_lines: Sequence[str] | None

    #: The original lines for the end revision of the interdiff.
    #:
    #: Version Added:
    #:     8.0
    _interdiff_interfilediff_orig_lines: Sequence[str] | None

    @deprecate_non_keyword_only_args(RemovedInReviewBoard90Warning)
    def __init__(
        self,
        request: HttpRequest,
        filediff: FileDiff,
        interfilediff: Optional[FileDiff] = None,
        force_interdiff: bool = False,
        base_filediff: Optional[FileDiff] = None,
        *,
        diff_settings: DiffSettings,
    ) -> None:
        """Initialize the DiffChunkGenerator.

        Version Changed:
            6.0:
            * Removed the old ``enable_syntax_highlighting`` argument.
            * Made ``diff_settings`` mandatory.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            filediff (reviewboard.diffviewer.models.filediff.FileDiff):
                The FileDiff to generate chunks for.

            interfilediff (reviewboard.diffviewer.models.filediff.FileDiff,
                           optional):
                If provided, the chunks will be generated for the differences
                between the result of applying ``filediff`` and the result of
                applying ``interfilediff``.

            force_interdiff (bool, optional):
                Whether or not to force an interdiff.

            base_filediff (reviewboard.diffviewer.models.filediff.FileDiff,
                           optional):
                An ancestor of ``filediff`` that we want to use as the base.
                Using this argument will result in the history between
                ``base_filediff`` and ``filediff`` being applied.

            diff_settings (reviewboard.diffviewer.settings.DiffSettings):
                The settings used to control the display of diffs.
        """
        assert filediff

        self.request = request
        self.diffset = filediff.diffset
        self.filediff = filediff
        self.interfilediff = interfilediff
        self.force_interdiff = force_interdiff
        self.repository = filediff.get_repository()
        self.tool = self.repository.get_scmtool()
        self.base_filediff = base_filediff

        self._interdiff_filediff_orig_lines = None
        self._interdiff_interfilediff_orig_lines = None

        if base_filediff:
            orig_filename = base_filediff.source_file
        else:
            orig_filename = filediff.source_file

        super().__init__(
            old=None,
            new=None,
            orig_filename=orig_filename,
            modified_filename=filediff.dest_file,
            encoding_list=self.repository.get_encoding_list(),
            diff_compat=filediff.diffset.diffcompat,
            diff_settings=diff_settings)

    def make_cache_key(self) -> str:
        """Return a new cache key for any generated chunks.

        Returns:
            str:
            The new cache key.
        """
        key: list[str] = []

        key.append('diff-sidebyside')

        if self.base_filediff is not None:
            key.append('base-%s' % self.base_filediff.pk)

        if not self.force_interdiff:
            key.append(str(self.filediff.pk))
        elif self.interfilediff:
            key.append('interdiff-%s-%s' % (self.filediff.pk,
                                            self.interfilediff.pk))
        else:
            key.append('interdiff-%s-none' % self.filediff.pk)

        key += [
            self.diff_settings.state_hash,
            get_language(),
        ]

        return '-'.join(key)

    def get_opcode_generator(self):
        """Return the DiffOpcodeGenerator used to generate diff opcodes."""
        diff = self.filediff.diff

        if self.interfilediff:
            interdiff = self.interfilediff.diff
        else:
            interdiff = None

        return get_diff_opcode_generator(
            self.differ,
            diff=diff,
            interdiff=interdiff,
            filediff_orig_lines=self._interdiff_filediff_orig_lines,
            interfilediff_orig_lines=self._interdiff_interfilediff_orig_lines,
            request=self.request,
            diff_settings=self.diff_settings,
        )

    def get_chunks(self):
        """Return the chunks for the given diff information.

        If the file is binary or is an added or deleted 0-length file, or if
        the file has moved with no additional changes, then an empty list of
        chunks will be returned.

        If there are chunks already computed in the cache, they will be
        yielded. Otherwise, new chunks will be generated, stored in cache,
        and yielded.
        """
        counts = self.filediff.get_line_counts()

        if (self.filediff.binary or
            self.filediff.source_revision == '' or
            ((self.filediff.is_new or self.filediff.deleted or
              self.filediff.moved or self.filediff.copied) and
             counts['raw_insert_count'] == 0 and
             counts['raw_delete_count'] == 0)):
            return

        cache_key = self.make_cache_key()

        for chunk in super(DiffChunkGenerator, self).get_chunks(cache_key):
            yield chunk

    def get_chunks_uncached(self):
        """Yield the list of chunks, bypassing the cache."""
        base_filediff = self.base_filediff
        filediff = self.filediff
        interfilediff = self.interfilediff
        request = self.request

        old, new = get_original_and_patched_files(
            filediff=filediff,
            request=request)

        old_encoding_list = get_filediff_encodings(filediff)
        new_encoding_list = old_encoding_list

        if base_filediff is not None:
            # The diff is against a commit that:
            #
            # 1. Follows the first commit in a series (the first won't have
            #    a base_commit/base_filediff that can be looked up)
            #
            # 2. Follows a commit that modifies this file, or is the base
            #    commit that modifies this file.
            #
            # We'll be diffing against the patched version of this commit's
            # version of the file.
            old = get_original_file(filediff=base_filediff,
                                    request=request)
            old = get_patched_file(source_data=old,
                                   filediff=base_filediff,
                                   request=request)
            old_encoding_list = get_filediff_encodings(base_filediff)
        elif filediff.commit_id:
            # This diff is against a commit, but no previous FileDiff
            # modifying this file could be found. As per the above comment,
            # this could end up being the very first commit in a series, or
            # it might not have been modified in the base commit or any
            # previous commit.
            #
            # We'll need to fetch the first ancestor of this file in the
            # commit history, if we can find one. We'll base the "old" version
            # of the file on the original version of this commit, meaning that
            # this commit and all modifications since will be shown as "new".
            # Basically, viewing the upstream of the file, before any commits.
            #
            # This should be safe because, without a base_filediff, there
            # should be no older commit containing modifications that we want
            # to diff against. This would be the first one, and we're using
            # its upstream changes.
            ancestors = filediff.get_ancestors(minimal=True)

            if ancestors:
                ancestor_filediff = ancestors[0]
                old = get_original_file(filediff=ancestor_filediff,
                                        request=request)
                old_encoding_list = get_filediff_encodings(ancestor_filediff)

        # Check whether we have a SHA256 checksum first. They were introduced
        # in Review Board 4.0, long after SHA1 checksums. If we already have
        # a SHA256 checksum, then we'll also have a SHA1 checksum, but the
        # inverse is not true.
        if filediff.orig_sha256 is None:
            if filediff.orig_sha1 is None:
                filediff.extra_data.update({
                    'orig_sha1': self._get_sha1(old),
                    'patched_sha1': self._get_sha1(new),
                })

            filediff.extra_data.update({
                'orig_sha256': get_sha256(old),
                'patched_sha256': get_sha256(new),
            })
            filediff.save(update_fields=['extra_data'])

        if interfilediff:
            # Keep a copy of the original content for use in interdiff
            # filtering.
            filediff_orig = old
            filediff_orig_encoding_list = old_encoding_list

            old = new
            old_encoding_list = new_encoding_list

            interdiff_orig = get_original_file(filediff=interfilediff,
                                               request=request)
            new = get_patched_file(source_data=interdiff_orig,
                                   filediff=interfilediff,
                                   request=request)
            new_encoding_list = get_filediff_encodings(interfilediff)

            # Check whether we have a SHA256 checksum first. They were
            # introduced in Review Board 4.0, long after SHA1 checksums. If we
            # already have a SHA256 checksum, then we'll also have a SHA1
            # checksum, but the inverse is not true.
            if interfilediff.orig_sha256 is None:
                if interfilediff.orig_sha1 is None:
                    interfilediff.extra_data.update({
                        'orig_sha1': self._get_sha1(interdiff_orig),
                        'patched_sha1': self._get_sha1(new),
                    })

                interfilediff.extra_data.update({
                    'orig_sha256': get_sha256(interdiff_orig),
                    'patched_sha256': get_sha256(new),
                })
                interfilediff.save(update_fields=['extra_data'])

            # Prepare the original versions of the files for both ends of the
            # interdiff so that we can pass them to get_opcode_generator()
            # later. This is used by the interdiff filtering algorithm.
            #
            # The patched sides of these will be available in the differ
            # as differ.a and differ.b.
            if filediff_orig is not None:
                self._interdiff_filediff_orig_lines = \
                    self.normalize_source_string(
                        filediff_orig, filediff_orig_encoding_list)[1]
            else:
                self._interdiff_filediff_orig_lines = []

            if interdiff_orig is not None:
                self._interdiff_interfilediff_orig_lines = \
                    self.normalize_source_string(
                        interdiff_orig, new_encoding_list)[1]
            else:
                self._interdiff_interfilediff_orig_lines = []
        elif self.force_interdiff:
            # Basically, revert the change.
            old, new = new, old
            old_encoding_list, new_encoding_list = \
                new_encoding_list, old_encoding_list

        timer_msg: str

        if interfilediff:
            timer_msg = (
                f'Generating diff chunks for interdiff ids '
                f'{filediff.pk}-{interfilediff.pk} ({filediff.source_file})'
            )
        else:
            timer_msg = (
                f'Generating diff chunks for filediff id {filediff.pk} '
                f'({filediff.source_file})'
            )

        with log_timed(timer_msg,
                       logger=logger,
                       request=request):
            yield from self.generate_chunks(
                old=old,
                new=new,
                old_encoding_list=old_encoding_list,
                new_encoding_list=new_encoding_list)

        if (not interfilediff and
            not self.base_filediff and
            not self.force_interdiff):
            insert_count = self.counts['insert']
            delete_count = self.counts['delete']
            replace_count = self.counts['replace']
            equal_count = self.counts['equal']

            filediff.set_line_counts(
                insert_count=insert_count,
                delete_count=delete_count,
                replace_count=replace_count,
                equal_count=equal_count,
                total_line_count=(insert_count + delete_count +
                                  replace_count + equal_count))

    def check_line_code_safety(self, orig_line, modified_line,
                               extra_state={}, **kwargs):
        """Check the safety of a line of code.

        This will run the original and modified line through all registered
        code safety checkers. If any checker produces warnings or errors,
        those will be associated with the line.

        This is a specialization of
        :py:meth:`RawDiffChunkGenerator.check_line_code_safety` that provides
        a ``repository`` key for each item to check, for use in the code
        safety checker.

        Version Added:
            5.0

        Args:
            orig_line (str):
                The original line to check.

            modified_line (str):
                The modiifed line to check.

            extra_state (dict, optional):
                Extra state to pass to the checker for the original or
                modified line content item. Used by subclasses to produce
                additional information that may be useful for some code safety
                checkers.

            **kwargs (dict, optional):
                Unused keyword arguments, for future expansion.

        Returns:
            list of tuple:
            A list of code safety results containing warnings or errors. Each
            item is a tuple containing:

            1. The registered checker ID.
            2. A dictionary with ``errors`` and/or ``warnings`` keys.
        """
        return super(DiffChunkGenerator, self).check_line_code_safety(
            orig_line=orig_line,
            modified_line=modified_line,
            extra_state=dict({
                'repository': self.repository,
            }, **extra_state),
            **kwargs)

    def normalize_path_for_display(self, filename):
        """Normalize a file path for display to the user.

        This uses the associated :py:class:`~reviewboard.scmtools.core.SCMTool`
        to normalize the filename.

        Args:
            filename (unicode):
                The filename to normalize.

        Returns:
            unicode:
            The normalized filename.
        """
        return self.tool.normalize_path_for_display(
            filename,
            extra_data=self.filediff.extra_data)

    def _get_sha1(self, content):
        """Return a SHA1 hash for the provided content.

        Args:
            content (bytes):
                The content to generate the hash for.

        Returns:
            unicode:
            The resulting hash.
        """
        return force_str(hashlib.sha1(content).hexdigest())


def compute_chunk_last_header(lines, numlines, meta, last_header=None):
    """Computes information for the displayed function/class headers.

    This will record the displayed headers, their line numbers, and expansion
    offsets relative to the header's collapsed line range.

    The last_header variable, if provided, will be modified, which is
    important when processing several chunks at once. It will also be
    returned as a convenience.
    """
    if last_header is None:
        last_header = [None, None]

    line = lines[0]

    for i, (linenum, header_key) in enumerate([(line[1], 'left_headers'),
                                               (line[4], 'right_headers')]):
        headers = meta[header_key]

        if headers:
            header = headers[-1]
            last_header[i] = {
                'line': header[0],
                'text': header[1].strip(),
            }

    return last_header


_generator: type[DiffChunkGenerator] = DiffChunkGenerator


def get_diff_chunk_generator_class() -> type[DiffChunkGenerator]:
    """Return the DiffChunkGenerator class used for generating chunks.

    Returns:
        type:
        The class for the DiffChunkGenerator to use.
    """
    return _generator


def set_diff_chunk_generator_class(
    renderer: type[DiffChunkGenerator],
) -> None:
    """Set the DiffChunkGenerator class used for generating chunks.

    Args:
        renderer (type):
            The class for the DiffChunkGenerator to use.
    """
    assert renderer

    globals()['_generator'] = renderer


def get_diff_chunk_generator(*args, **kwargs) -> DiffChunkGenerator:
    """Return a DiffChunkGenerator instance used for generating chunks.

    Args:
        *args (tuple):
            Positional arguments to pass to the DiffChunkGenerator.

        **kwargs (dict):
            Keyword arguments to pass to the DiffChunkGenerator.

    Returns:
        DiffChunkGenerator:
        The chunk generator instance.
    """
    return _generator(*args, **kwargs)



"""X.509 authentication backend."""

from __future__ import annotations

import logging
import re
from typing import Optional, TYPE_CHECKING

from django.conf import settings
from django.contrib.auth.models import User
from django.utils.translation import gettext_lazy as _

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.forms.auth import X509SettingsForm

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


class X509Backend(BaseAuthBackend):
    """Authenticate a user from a X.509 client certificate.

    The certificate is passed in by the browser. This backend relies on the
    X509AuthMiddleware to extract a username field from the client certificate.
    """

    backend_id = 'x509'
    name = _('X.509 Public Key')
    settings_form = X509SettingsForm
    supports_change_password = True

    def authenticate(
        self,
        request: Optional[HttpRequest] = None,
        *,
        x509_field: str = '',
        **kwargs,
    ) -> Optional[User]:
        """Authenticate the user.

        This will extract the username from the provided certificate and return
        the appropriate User object.

        Version Changed:
            6.0:
            * ``request`` is now optional.
            * ``username`` and ``password`` are technically optional, to
              aid in consistency for type hints, but will result in a ``None``
              result.

        Version Changed:
            4.0:
            The ``request`` argument is now mandatory as the first positional
            argument, as per requirements in Django.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the caller. This may be ``None``.

            x509_field (str, optional):
                The value of the field containing the username.

            **kwargs (dict, unused):
                Additional keyword arguments supplied by the caller.

        Returns:
            django.contrib.auth.models.User:
            The authenticated user, or ``None`` if the user could not be
            authenticated for any reason.
        """
        if not x509_field:
            # This may be an authentication request for a backend expecting
            # different arguments.
            return None

        username = self.clean_username(x509_field)
        return self.get_or_create_user(username=username,
                                       request=request)

    def clean_username(
        self,
        username: str,
    ) -> str:
        """Validate the 'username' field.

        This checks to make sure that the contents of the username field are
        valid for X509 authentication.

        Args:
            username (str):
                The username to validate.

        Returns:
            str:
            The resulting username.
        """
        username = username.strip()

        x509_username_regex = getattr(settings, 'X509_USERNAME_REGEX', None)

        if x509_username_regex:
            try:
                m = re.match(x509_username_regex, username)

                if m:
                    username = m.group(1)
                else:
                    logger.warning("X509Backend: username '%s' didn't match "
                                   "regex.", username)
            except re.error as e:
                logger.error('X509Backend: Invalid regex specified: %s',
                             e, exc_info=True)

        return username

    def get_or_create_user(
        self,
        username: str,
        request: Optional[HttpRequest] = None,
    ) -> Optional[User]:
        """Return an existing user or create one if it doesn't exist.

        This does not authenticate the user.

        If the user does not exist in the database, but does in the backend,
        its information will be stored in the database for later lookup.

        Args:
            username (str):
                The username to fetch or create.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            django.contrib.auth.models.User:
            The resulting user, or ``None`` if one could not be found.
        """
        user: Optional[User] = None
        username = username.strip()

        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            # TODO Add the ability to get the first and last names in a
            #      configurable manner; not all X.509 certificates will have
            #      the same format.
            if getattr(settings, 'X509_AUTOCREATE_USERS', False):
                user = User(username=username, password='')
                user.is_staff = False
                user.is_superuser = False
                user.set_unusable_password()
                user.save()

        return user



"""Unit tests for reviewboard.reviews.models.status_update.StatusUpdate."""

from __future__ import annotations

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser, Permission, User
from django.db.models import Q
from djblets.testing.decorators import add_fixtures

from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.models.base_comment import BaseComment
from reviewboard.reviews.models.review_request import fetch_issue_counts
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class StatusUpdateTests(TestCase):
    """Unit tests for reviewboard.reviews.models.base_comment.StatusUpdate."""

    fixtures = ['test_users', 'test_scmtools']

    def test_is_mutable_by_with_anonymous(self):
        """Testing StatusUpdate.is_mutable_by with anonymous user"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(review_request)

        self.assertFalse(status_update.is_mutable_by(AnonymousUser()))

    def test_is_mutable_by_with_owner(self):
        """Testing StatusUpdate.is_mutable_by with owner"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(review_request)

        self.assertTrue(status_update.is_mutable_by(status_update.user))

    def test_is_mutable_by_with_other_user(self):
        """Testing StatusUpdate.is_mutable_by with other user"""
        other_user = User.objects.create(username='other-user')
        review_request = self.create_review_request()
        status_update = self.create_status_update(review_request)

        self.assertFalse(status_update.is_mutable_by(other_user))

    def test_is_mutable_by_with_other_user_and_can_change_status_perm(self):
        """Testing StatusUpdate.is_mutable_by with other user with
        change_statusupdate permission
        """
        other_user = User.objects.create(username='other-user')
        other_user.user_permissions.add(
            Permission.objects.get(codename='change_statusupdate'))

        review_request = self.create_review_request()
        status_update = self.create_status_update(review_request)

        self.assertTrue(status_update.is_mutable_by(other_user))

    @add_fixtures(['test_site'])
    def test_is_mutable_by_with_other_user_with_perm_same_local_site(self):
        """Testing StatusUpdate.is_mutable_by with other user on same
        LocalSite with change_statusupdate permission
        """
        review_request = self.create_review_request(with_local_site=True)
        status_update = self.create_status_update(review_request)

        other_user = User.objects.create(username='other-user')

        site = review_request.local_site
        site.users.add(other_user)

        site_profile = other_user.get_site_profile(site)
        site_profile.permissions = {
            'reviews.change_statusupdate': True,
        }
        site_profile.save(update_fields=('permissions',))

        self.assertTrue(status_update.is_mutable_by(other_user))

    def test_drop_open_issues_with_no_review(self):
        """Testing StatusUpdate.drop_open_issues with no associated Review"""
        review_request = self.create_review_request(publish=True)
        status_update = self.create_status_update(review_request, review=None)
        status_update.drop_open_issues()

    def test_drop_open_issues_multiple_issue_status(self):
        """Testing StatusUpdate.drop_open_issues with multiple issues
        statuses
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, publish=True)

        self.create_diff_comment(review, filediff, issue_opened=True)
        self.create_diff_comment(review, filediff, issue_opened=True,
                                 issue_status=BaseComment.RESOLVED)
        self.create_diff_comment(review, filediff, issue_opened=True,
                                 issue_status=BaseComment.DROPPED)

        status_update = self.create_status_update(review_request,
                                                  review=review)

        status_update.drop_open_issues()

        comments = list(review.comments.all().order_by('id'))
        self.assertEqual(len(comments), 3)
        self.assertEqual(comments[0].issue_status, BaseComment.DROPPED)
        self.assertEqual(comments[1].issue_status, BaseComment.RESOLVED)
        self.assertEqual(comments[2].issue_status, BaseComment.DROPPED)

    def test_drop_open_issues_updates_timestamps(self):
        """Testing StatusUpdate.drop_open_issues updates timestamps"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, publish=True)
        open_comment = self.create_diff_comment(
            review, filediff, issue_opened=True)
        resolved_comment = self.create_diff_comment(
            review, filediff, issue_opened=True,
            issue_status=BaseComment.RESOLVED)
        status_update = self.create_status_update(review_request,
                                                  review=review)

        open_comment_timestamp = open_comment.timestamp
        resolved_comment_timestamp = resolved_comment.timestamp
        review_request_timestamp = \
            review_request.last_review_activity_timestamp

        status_update.drop_open_issues()

        comments = list(review.comments.all().order_by('id'))
        self.assertEqual(len(comments), 2)
        self.assertGreater(comments[0].timestamp, open_comment_timestamp)
        self.assertEqual(comments[1].timestamp, resolved_comment_timestamp)

        self.assertGreater(review_request.last_review_activity_timestamp,
                           review_request_timestamp)

    def test_drop_open_issues_updates_issue_counts(self):
        """Testing StatusUpdate.drop_open_issues updates issue counts"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, publish=True)

        self.create_diff_comment(review, filediff, issue_opened=True)

        status_update = self.create_status_update(review_request,
                                                  review=review)

        issues = fetch_issue_counts(review_request)
        self.assertEqual(issues[BaseComment.OPEN], 1)
        self.assertEqual(issues[BaseComment.DROPPED], 0)
        self.assertEqual(issues[BaseComment.RESOLVED], 0)

        status_update.drop_open_issues()

        issues = fetch_issue_counts(review_request)
        self.assertEqual(issues[BaseComment.OPEN], 0)
        self.assertEqual(issues[BaseComment.DROPPED], 1)
        self.assertEqual(issues[BaseComment.RESOLVED], 0)

    def test_get_integration_config(self) -> None:
        """Testing StatusUpdate.integration_config getter"""
        # Pre-fetch the LocalSite cache.
        LocalSite.objects.has_local_sites()

        review_request = self.create_review_request()
        config = IntegrationConfig.objects.create(name='my-config')
        status_update = self.create_status_update(
            review_request,
            extra_data={
                '__integration_config_id': config.pk,
            })

        queries = [
            {
                'model': IntegrationConfig,
                'where': Q(pk=config.pk),
            },
        ]

        with assert_queries(queries):
            config2 = status_update.integration_config

        self.assertEqual(config, config2)

        # Check that the configuration is cached.
        with self.assertNumQueries(0):
            status_update.integration_config

    @add_fixtures(['test_site'])
    def test_get_integration_config_with_local_site(self) -> None:
        """Testing StatusUpdate.integration_config getter with Local Site"""
        # Pre-fetch the LocalSite cache.
        LocalSite.objects.has_local_sites()

        local_site = self.get_local_site(self.local_site_name)
        review_request = self.create_review_request(local_site=local_site)
        config = IntegrationConfig.objects.create(name='my-config',
                                                  local_site=local_site)
        status_update = self.create_status_update(
            review_request,
            extra_data={
                '__integration_config_id': config.pk,
            })

        queries = [
            {
                'model': IntegrationConfig,
                'where': Q(pk=config.pk) & Q(local_site=local_site.pk),
            },
        ]

        with assert_queries(queries):
            config2 = status_update.integration_config

        self.assertEqual(config, config2)

        # Check that the configuration is cached.
        with self.assertNumQueries(0):
            status_update.integration_config

    @add_fixtures(['test_site'])
    def test_get_integration_config_with_bad_local_site(self) -> None:
        """Testing StatusUpdate.integration_config getter with Local Site
        mismatch
        """
        # Pre-fetch the LocalSite cache.
        LocalSite.objects.has_local_sites()

        local_site = self.get_local_site(self.local_site_name)
        review_request = self.create_review_request(local_site=local_site)
        config = IntegrationConfig.objects.create(name='my-config')
        status_update = self.create_status_update(
            review_request,
            extra_data={
                '__integration_config_id': config.pk,
            })

        queries = [
            {
                'model': IntegrationConfig,
                'where': Q(pk=config.pk) & Q(local_site=local_site.pk),
            },
        ]

        with assert_queries(queries):
            config2 = status_update.integration_config

        self.assertIsNone(config2)

        # Check that the result is cached.
        with self.assertNumQueries(0):
            status_update.integration_config

    def test_get_integration_config_with_missing(self) -> None:
        """Testing StatusUpdate.integration_config getter with missing
        IntegrationConfig
        """
        # Pre-fetch the LocalSite cache.
        LocalSite.objects.has_local_sites()

        review_request = self.create_review_request()
        status_update = self.create_status_update(
            review_request,
            extra_data={
                '__integration_config_id': 123,
            })

        queries = [
            {
                'model': IntegrationConfig,
                'where': Q(pk=123),
            },
        ]

        with assert_queries(queries):
            config2 = status_update.integration_config

        self.assertIsNone(config2)

        # Check that the result is cached.
        with self.assertNumQueries(0):
            status_update.integration_config

    def test_set_integration_config(self) -> None:
        """Testing StatusUpdate.integration_config setter"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(review_request)

        self.assertEqual(status_update.extra_data, {})

        # Set the configuration.
        config = IntegrationConfig.objects.create(name='my-config')
        status_update.integration_config = config

        self.assertEqual(status_update.extra_data, {
            '__integration_config_id': config.pk,
        })
        self.assertEqual(status_update.integration_config, config)

    def test_set_integration_config_with_none(self) -> None:
        """Testing StatusUpdate.integration_config setter with None"""
        review_request = self.create_review_request()
        status_update = self.create_status_update(review_request)

        self.assertEqual(status_update.extra_data, {})

        # Set the initial configuration.
        config = IntegrationConfig.objects.create(name='my-config')
        status_update.integration_config = config

        self.assertEqual(status_update.extra_data, {
            '__integration_config_id': config.pk,
        })
        self.assertEqual(status_update.integration_config, config)

        # Set it back to None.
        status_update.integration_config = None

        self.assertEqual(status_update.extra_data, {})
        self.assertIsNone(status_update.integration_config)

    def test_set_integration_config_with_bad_local_site(self) -> None:
        """Testing StatusUpdate.integration_config setter with LocalSite
        mismatch
        """
        review_request = self.create_review_request()
        status_update = self.create_status_update(review_request)

        self.assertEqual(status_update.extra_data, {})

        # Set the initial configuration.
        local_site = self.create_local_site(name='test-site-1')
        config = IntegrationConfig.objects.create(name='my-config',
                                                  local_site=local_site)

        message = (
            'The integration configuration and Status Update must have the '
            'same Local Site.'
        )

        with self.assertRaisesMessage(ValueError, message):
            status_update.integration_config = config

        self.assertEqual(status_update.extra_data, {})
        self.assertIsNone(status_update.integration_config)



"""Unit tests for reviewboard.accounts.views.ClientLoginView.

Version Added:
    5.0.5
"""

from __future__ import annotations

import datetime
from typing import Optional
from urllib.parse import quote

import kgb
from django.contrib.auth.models import User
from django.template import Context
from django.utils import timezone
from django.utils.html import escape
from djblets.siteconfig.models import SiteConfiguration
from djblets.webapi.errors import WebAPITokenGenerationError

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase
from reviewboard.webapi.models import WebAPIToken


class ClientLoginViewTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.accounts.views.ClientLoginView.

    Version Added:
        5.0.5
    """

    fixtures = ['test_users']

    def test_get(self) -> None:
        """Testing ClientLoginView GET builds a payload containing
        authentication data when the user is logged in and the client web
        login flow is enabled
        """
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(
            timezone.make_aware(datetime.datetime(2023, 5, 20))))

        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        self._assert_context_equals(
            rsp.context,
            client_allowed=True,
            client_name='TestClient',
            client_url='http://localhost:1234/test/',
            username='doc',
            check_payload_token=True,
            token_expires=timezone.make_aware(datetime.datetime(2023, 5, 25)))

    def test_get_siteconfig_not_set(self) -> None:
        """Testing ClientLoginView GET still builds a payload containing
        authentication data when the client web login and the client
        token expiration settings are missing from the site configuration
        """
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(
            timezone.make_aware(datetime.datetime(2023, 5, 20))))

        siteconfig = SiteConfiguration.objects.get_current()
        del siteconfig.settings['client_web_login']
        del siteconfig.settings['client_token_expiration']

        self.assertNotIn('client_web_login', siteconfig.settings)
        self.assertNotIn('client_token_expiration', siteconfig.settings)

        self.client.login(username='doc', password='doc')

        rsp = self.client.get(
            local_site_reverse('client-login'),
            {
                'client-name': 'TestClient',
                'client-url': 'http://localhost:1234/test/',
            })

        self._assert_context_equals(
            rsp.context,
            client_allowed=True,
            client_name='TestClient',
            client_url='http://localhost:1234/test/',
            username='doc',
            check_payload_token=True,
            token_expires=timezone.make_aware(datetime.datetime(2024, 5, 19)))

    def test_get_with_redirect(self) -> None:
        """Testing ClientLoginView GET with a redirect URL"""
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(
            timezone.make_aware(datetime.datetime(2023, 5, 20))))

        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                    'next': 'http://localhost:1234/page?foo=1',
                })

        self._assert_context_equals(
            rsp.context,
            client_allowed=True,
            client_name='TestClient',
            client_url='http://localhost:1234/test/',
            redirect_to='http%3A//localhost%3A1234/page%3Ffoo%3D1',
            username='doc',
            check_payload_token=True,
            token_expires=timezone.make_aware(datetime.datetime(2023, 5, 25)))

    def test_get_with_unsafe_redirect(self) -> None:
        """Testing ClientLoginView GET with an unsafe redirect URL"""
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(
            timezone.make_aware(datetime.datetime(2023, 5, 20))))

        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                    'next': 'http://unsafe-site/page?foo=1',
                })

        self._assert_context_equals(
            rsp.context,
            client_allowed=True,
            client_name='TestClient',
            client_url='http://localhost:1234/test/',
            redirect_to='',
            username='doc',
            check_payload_token=True,
            token_expires=timezone.make_aware(datetime.datetime(2023, 5, 25)))

    def test_get_with_token_generation_error(self) -> None:
        """Testing ClientLoginView GET with an API token generation error"""
        self.spy_on(WebAPIToken.objects.get_or_create_client_token,
                    op=kgb.SpyOpRaise(WebAPITokenGenerationError('fail')))

        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        tokens = WebAPIToken.objects.filter(
            user=User.objects.get(username='doc'))

        self.assertEqual(len(tokens), 0)
        self._assert_context_equals(
            rsp.context,
            client_allowed=True,
            client_name='TestClient',
            client_url='http://localhost:1234/test/',
            error='Failed to generate a unique API token for authentication. '
                  'Please reload the page to try again.',
            username='doc')

    def test_get_unauthenticated(self) -> None:
        """Testing ClientLoginView GET redirects to the login page when
        a user is not logged in
        """
        settings = {
            'client_web_login': True,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        self.assertEqual(rsp.status_code, 302)

    def test_get_with_unsafe_client_url(self) -> None:
        """Testing ClientLoginView GET with an unsafe client url"""

        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            with self.assertLogs() as logs:
                rsp = self.client.get(
                    local_site_reverse('client-login'),
                    {
                        'client-name': 'TestClient',
                        'client-url': 'http://unsafe-url.com',
                    })

        tokens = WebAPIToken.objects.filter(
            user=User.objects.get(username='doc'))

        self.assertEqual(len(tokens), 0)
        self._assert_context_equals(
            rsp.context,
            client_allowed=False,
            client_name='TestClient',
            client_url='http://unsafe-url.com',
            username='doc')
        self.assertEqual(
            logs.records[0].getMessage(),
            ('Blocking an attempt to send authentication info '
             'to unsafe URL http://unsafe-url.com'))

    def test_get_with_client_url_no_port(self) -> None:
        """Testing ClientLoginView GET with a client url that has no port
        specified
        """
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(
            timezone.make_aware(datetime.datetime(2023, 5, 20))))

        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost',
                })

        self._assert_context_equals(
            rsp.context,
            client_allowed=True,
            client_name='TestClient',
            client_url='http://localhost',
            username='doc',
            check_payload_token=True,
            token_expires=timezone.make_aware(datetime.datetime(2023, 5, 25)))

    def test_get_with_client_web_login_false(self) -> None:
        """Testing ClientLoginView GET with the client web login flow
        disabled
        """
        settings = {
            'client_web_login': False,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        self.assertEqual(rsp.status_code, 404)

    def test_post(self) -> None:
        """Testing ClientLoginView POST"""
        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
            'client_token_expiration': 5
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.post(
                local_site_reverse('client-login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        self.assertEqual(rsp.status_code, 405)

    def _assert_context_equals(
        self,
        context: Context,
        client_allowed: bool,
        client_name: str,
        client_url: str,
        username: str,
        check_payload_token: Optional[bool] = False,
        error: Optional[str] = '',
        redirect_to: Optional[str] = '',
        token_expires: Optional[datetime.datetime] = None,
    ) -> None:
        """Assert that the context and JS view data matches the given values.

        Args:
            context (django.template.Context):
                The context dictionary to be tested.

            client_allowed (bool):
                The expected value for the client_allowed.

            client_name (str):
                The expected value for the client_name.

            client_url (str):
                The expected value for the client_url.

            username (str):
                The expected value for the username.

            check_payload_token (bool, optional):
                Whether to check for an API token in the payload.

            error (str, optional):
                The expected value for the error.

            redirect_to (str, optional):
                The expected value for the redirect_to.

            token_expires (datetime.datetime, optional):
                The expected value for the API token expiration.

        Raises:
            AssertionError:
                The context did not match the given values.
        """
        js_view_data = context['js_view_data']
        payload = js_view_data['payload']

        self.assertEqual(context['client_allowed'], client_allowed)
        self.assertEqual(context['client_name'], client_name)
        self.assertEqual(context['client_url'], client_url)
        self.assertEqual(context['username'], username)

        if error:
            self.assertEqual(context['error'], error)
        else:
            self.assertNotIn('error', context)

        self.assertEqual(js_view_data['clientName'], escape(client_name))
        self.assertEqual(js_view_data['clientURL'], quote(client_url))
        self.assertEqual(js_view_data['username'], username)
        self.assertEqual(js_view_data['redirectTo'], redirect_to)

        if check_payload_token:
            token = WebAPIToken.objects.get(token=payload['api_token'])
            self.assertEqual(token.user.username, username)
            self.assertEqual(token.expires, token_expires)
            self.assertEqual(token.extra_data['client_name'], client_name)
        else:
            self.assertEqual(payload, {})



"""Root file attachment comments API resource.

Version Added:
    5.0
"""

from django.contrib.auth.models import User
from django.db.models import Q
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import webapi_request_fields
from djblets.webapi.fields import (BooleanFieldType,
                                   DateTimeFieldType,
                                   StringFieldType)

from reviewboard.attachments.models import FileAttachment
from reviewboard.reviews.models import FileAttachmentComment
from reviewboard.webapi.resources.base_file_attachment_comment import \
    BaseFileAttachmentCommentResource


class RootFileAttachmentCommentResource(BaseFileAttachmentCommentResource):
    """Provide information on file attachment comments.

    This is a top level endpoint that allows you to list and query all
    file attachment comments in the system, across different review requests.

    Version Changed:
        5.0.4:
        This was formally named ``file_attachment_comments`` on the root
        resource, but had to be renamed to ``all_file_attachment_comments`` in
        order to work around a conflict with URI templates.

    Version Added:
        5.0
    """

    added_in = '5.0'
    allowed_methods = ('GET',)
    model = FileAttachmentComment
    link_name = 'all_file_attachment_comments'
    uri_template_name = 'all_file_attachment_comment'
    uri_object_key = None

    def get_queryset(self, request, is_list=False, *args, **kwargs):
        """Return a queryset for FileAttachmentComment models.

        By default, this returns all file attachment comments that
        are accessible to the requester.

        The queryset can be further filtered by one or more arguments
        in the URL. These are listed in :py:meth:`get_list`.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            is_list (bool, unused):
                Whether the coming HTTP request is request for list resources.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for FileAttachmentComment models.
        """
        local_site = request.local_site
        q = Q()

        if 'file-attachment-id' in request.GET:
            q &= Q(file_attachment=request.GET.get('file-attachment-id'))

        if 'file-name' in request.GET:
            file_attachment = list((
                FileAttachment.objects
                .filter(orig_filename=request.GET.get('file-name'))
                .values_list('pk', flat=True)
            ))

            if file_attachment:
                q &= Q(file_attachment=file_attachment[0])
            else:
                return self.model.objects.none()

        if 'is-reply' in request.GET:
            val = request.GET.get('is-reply')
            q &= Q(reply_to_id__isnull=(not val))

        if 'last-updated-from' in request.GET:
            q &= Q(timestamp__gte=request.GET.get('last-updated-from'))

        if 'last-updated-to' in request.GET:
            q &= Q(timestamp__lt=request.GET.get('last-updated-to'))

        if 'review-id' in request.GET:
            q &= Q(review=request.GET.get('review-id'))

        if 'review-request-id' in request.GET:
            review_request_id = request.GET.get('review-request-id')

            if local_site is None:
                q &= Q(review__review_request=review_request_id)
            else:
                q &= (Q(review__review_request__local_id=review_request_id) &
                      Q(review__review_request__local_site=local_site))

        if 'user' in request.GET:
            user = list((
                User.objects
                .filter(username=request.GET.get('user'))
                .values_list('pk', flat=True)
            ))

            if user:
                q &= Q(review__user=user[0])
            else:
                return self.model.objects.none()

        return self.model.objects.accessible(request.user,
                                             extra_query=q,
                                             local_site=local_site)

    @webapi_request_fields(
        optional={
            'file-attachment-id': {
                'type': StringFieldType,
                'description': 'The ID of the file attachment that the '
                               'comments must belong to.',
            },
            'file-name': {
                'type': StringFieldType,
                'description': 'The filename of the file attachment name '
                               'that the comments must belong to.',
            },
            'is-reply': {
                'type': BooleanFieldType,
                'description': 'Determine whether to return file attachment '
                               'comments that are replies or not.',
            },
            'last-updated-to': {
                'type': DateTimeFieldType,
                'description': "The date/time that all file attachment "
                               "comments must be last updated before. This is "
                               "compared against the file attachment "
                               "comment's ``timestamp`` field. This must be "
                               "a valid :term:`date/time format`.",
            },
            'last-updated-from': {
                'type': DateTimeFieldType,
                'description': "The earliest date/time the file attachment "
                               "comments could be last updated. This is "
                               "compared against the file attachment "
                               "comment's ``timestamp`` field. This must "
                               "be a valid :term:`date/time format`.",
            },
            'review-id': {
                'type': StringFieldType,
                'description': 'The ID of the review that the file attachment '
                               'comments must belong to.',
            },
            'review-request-id': {
                'type': StringFieldType,
                'description': 'The ID of the review request that the file '
                               'attachment comments must belong to.',
            },
            'user': {
                'type': StringFieldType,
                'description': 'The username of the user that created the '
                               'file attachment comments.',
            },
        },
        allow_unknown=True
    )
    @augment_method_from(BaseFileAttachmentCommentResource)
    def get_list(self, *args, **kwargs):
        """Return the list of file attachment comments."""
        pass


root_file_attachment_comment_resource = RootFileAttachmentCommentResource()



"""Views for the Forgejo hosting service.

Version Added:
    8.0
"""

from __future__ import annotations

import hashlib
import hmac
import logging
from collections import defaultdict
from typing import List, TYPE_CHECKING

from django.http import HttpResponse, HttpResponseBadRequest
from django.views import View
from pydantic import BaseModel, ValidationError

from reviewboard.admin.server import get_server_url
from reviewboard.hostingsvcs.hook_utils import (
    close_all_review_requests,
    get_git_branch_name,
    get_repository_for_hook,
    get_review_request_id,
)

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


class PushWebHookPayload(BaseModel):
    """Data for the push event WebHook payload.

    Version Added:
        8.0
    """

    class _Commit(BaseModel):
        id: str
        message: str

    commits: List[_Commit]
    ref: str


class WebHookView(View):
    """WebHook handler for Forgejo.

    Version Added:
        8.0
    """

    def post(
        self,
        request: HttpRequest,
        *,
        local_site_name: (str | None) = None,
        repository_id: int,
        hosting_service_id: str,
    ) -> HttpResponse:
        """Handle a POST request to the view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the Forgejo server.

            local_site_name (str):
                The Local Site name, if available.

            repository_id (int):
                The ID of the repository.

            hosting_service_id (str):
                The name of the hosting service.

        Returns:
            django.http.HttpResponse:
            The response to send back to the server.
        """
        hook_event = request.META.get('HTTP_X_FORGEJO_EVENT')

        if hook_event != 'push':
            return HttpResponseBadRequest(
                'Only the "push" event is supported.')

        repository = get_repository_for_hook(
            repository_id=repository_id,
            hosting_service_id=hosting_service_id,
            local_site_name=local_site_name)

        signature = request.META.get('HTTP_X_FORGEJO_SIGNATURE')

        if not signature:
            logger.error('Forgejo WebHook: received POST with a missing '
                         'signature (repository %r).',
                         repository)

            return HttpResponseBadRequest(
                'X-Forgejo-Signature header was missing or blank.')

        m = hmac.new(repository.get_or_create_hooks_uuid().encode(),
                     request.body, hashlib.sha256)

        if m.hexdigest() != signature:
            logger.error('Forgejo WebHook: received incorrect signature '
                         '(repository %r).',
                         repository)

            return HttpResponseBadRequest(
                'Payload signature could not be validated.')

        try:
            payload = PushWebHookPayload.model_validate_json(request.body)
        except ValidationError as e:
            logger.error('Forgejo WebHook: Could not validate payload content '
                         '(repository %r): %s',
                         repository, e)

            return HttpResponseBadRequest(
                'Payload could not be parsed, see Review Board log for '
                'details.')

        server_url = get_server_url(request=request)
        review_request_id_to_commits: defaultdict[int | None, list[str]] = \
            defaultdict(list)
        branch_name = get_git_branch_name(payload.ref)

        for commit in payload.commits:
            review_request_id = get_review_request_id(
                commit_message=commit.message,
                server_url=server_url,
                commit_id=commit.id,
                repository=repository)
            assert review_request_id is not None

            review_request_id_to_commits[review_request_id].append(
                f'{branch_name} ({commit.id[:7]})')

        if review_request_id_to_commits:
            logger.info(
                'Forgejo WebHook: Closing review requests %s',
                ', '.join(str(key) for key in review_request_id_to_commits))

            close_all_review_requests(
                review_request_id_to_commits=review_request_id_to_commits,
                local_site_name=local_site_name,
                repository=repository,
                hosting_service_id=hosting_service_id)

        return HttpResponse()



"""HTTP Digest authentication backend."""

from __future__ import annotations

import hashlib
import logging
from typing import Optional, TYPE_CHECKING

from django.conf import settings
from django.contrib.auth.models import User
from django.utils.translation import gettext_lazy as _

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.forms.auth import HTTPBasicSettingsForm

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


class HTTPDigestBackend(BaseAuthBackend):
    """Authenticate against a user in a digest password file.

    This is controlled by the following Django settings:

    .. setting:: DIGEST_FILE_LOCATION

    ``DIGEST_FILE_LOCATION``:
        The local file path on the server containing an HTTP password
        (:file:`htpasswd`) file.

        This is ``auth_digest_file_location`` in the site configuration.


    .. setting:: DIGEST_REALM

    ``DIGEST_REALM``:
        The HTTP realm users will be authenticated into.

        This is ``auth_digest_realm`` in the site configuration.
    """

    backend_id = 'digest'
    name = _('HTTP Digest Authentication')
    settings_form = HTTPBasicSettingsForm
    login_instructions = _('Use your standard username and password.')

    def authenticate(
        self,
        request: Optional[HttpRequest] = None,
        *,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ) -> Optional[User]:
        """Authenticate a user against the HTTP password file.

        This will attempt to authenticate the user against the digest password
        file. If the username and password are valid, a user will be returned,
        and added to the database if it doesn't already exist.

        Version Changed:
            6.0:
            * ``request`` is now optional.
            * ``username`` and ``password`` are technically optional, to
              aid in consistency for type hints, but will result in a ``None``
              result.

        Version Changed:
            4.0:
            The ``request`` argument is now mandatory as the first positional
            argument, as per requirements in Django.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the caller. This may be ``None``.

            username (str):
                The username to authenticate.

            password (str):
                The user's password.

            **kwargs (dict, unused):
                Additional keyword arguments passed by the caller.

        Returns:
            django.contrib.auth.models.User:
            The authenticated user, or ``None`` if the user could not be
            authenticated for any reason.
        """
        if not username or not password:
            # This may be an authentication request for a backend expecting
            # different arguments.
            return None

        username = username.strip()

        filename = settings.DIGEST_FILE_LOCATION
        digest_text = '%s:%s:%s' % (username, settings.DIGEST_REALM, password)
        digest_password = hashlib.md5(digest_text.encode('utf-8')).hexdigest()

        try:
            with open(filename, 'r') as fp:
                for line_no, line in enumerate(fp):
                    try:
                        user, realm, passwd = line.strip().split(':')

                        if user == username and passwd == digest_password:
                            return self.get_or_create_user(username=username,
                                                           request=request)
                    except ValueError as e:
                        logger.exception('Error parsing HTTP Digest password '
                                         'file "%s" at line %d: %s',
                                         filename, line_no, e,
                                         extra={'request': request})
                        break
        except IOError as e:
            logger.exception('Could not open the HTTP Digest password '
                             'file "%s": %s',
                             filename, e,
                             extra={'request': request})

        return None

    def get_or_create_user(
        self,
        username: str,
        request: Optional[HttpRequest] = None,
    ) -> Optional[User]:
        """Return an existing user or create one if it doesn't exist.

        This does not authenticate the user.

        If the user does not exist in the database, but does in the HTTP
        password file, its information will be stored in the database for later
        lookup.

        Args:
            username (str):
                The name of the user to look up or create.

            request (django.http.HttpRequest, unused):
                The HTTP request from the client. This is unused.

        Returns:
            django.contrib.auth.models.User:
            The resulting user, or ``None`` if one could not be found.
        """
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            user = User.objects.create_user(username=username)

        return user



"""Testing utilities for API unit tests.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Optional, Set, TYPE_CHECKING, Type, Union

from django.contrib.auth.models import AnonymousUser, User
from django.db.models import Count, Q
from oauth2_provider.models import AccessToken

from reviewboard.accounts.testing.queries import get_user_profile_equeries
from reviewboard.oauth.models import Application
from reviewboard.site.testing.queries import (
    get_check_local_site_access_equeries,
)
from reviewboard.testing.queries.http import get_http_request_user_equeries
from reviewboard.webapi.models import WebAPIToken

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries
    from django.db.models import Model

    from reviewboard.site.models import LocalSite
    from reviewboard.testing.queries.base import ExpectedQResult


def get_webapi_token_equeries(
    *,
    user: User,
    webapi_token: WebAPIToken,
) -> ExpectedQueries:
    """Return expected queries for fetching API tokens for a user.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.User):
            The user that owns the API token.

        webapi_token (reviewboard.webapi.models.WebAPIToken):
            The API token being fetched.

    Returns:
        list of dict:
        The list of expected queries.
    """
    return [
        {
            '__note__': 'Fetch the API token for the request',
            'model': WebAPIToken,
            'where': (Q(pk=webapi_token.pk) &
                      Q(user=user)),
        },
    ]


def get_webapi_request_start_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: Optional[LocalSite] = None,
    webapi_token: Optional[WebAPIToken] = None,
    oauth2_access_token: Optional[AccessToken] = None,
    oauth2_application: Optional[Application] = None,
) -> ExpectedQueries:
    """Return expected queries for the start of an API HTTP request.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        checks_local_site_access (bool, optional):
            Whether the requested view uses the
            :py:func:`~reviewboard.site.decorators.check_local_site_access`
            decorator.

    Returns:
        list of dict:
        The list of expected queries.
    """
    assert (oauth2_access_token is None) is (oauth2_application is None)

    equeries: ExpectedQueries = []

    if user.is_authenticated:
        assert isinstance(user, User)

        equeries += get_http_request_user_equeries(user=user)
        equeries += get_user_profile_equeries(user=user)

        if webapi_token is not None:
            equeries += get_webapi_token_equeries(
                user=user,
                webapi_token=webapi_token)

        if oauth2_access_token is not None:
            assert oauth2_application is not None

            equeries += [
                {
                    '__note__': 'Fetch the OAuth2 token for the request',
                    'model': AccessToken,
                    'where': Q(pk=oauth2_access_token.pk),
                },
                {
                    '__note__': 'Fetch the OAuth2 application for the token',
                    'model': Application,
                    'where': Q(id=oauth2_application.pk),
                },
            ]

    if local_site is not None:
        equeries += get_check_local_site_access_equeries(
            user=user,
            local_site=local_site)

    return equeries


def get_webapi_response_start_equeries(
    *,
    model: Type[Model],
    items_q_result: ExpectedQResult,
    items_distinct: bool = False,
    items_select_related: Set[str] = set(),
) -> ExpectedQueries:
    """Return expected queries for the start of an API payload response.

    Version Added:
        5.0.7

    Args:
        model (type):
            The type of model backing the resource.

        items_q_result (dict):
            An expected Q result representing the queries for items and
            item counts.

        items_distinct (bool, optional):
            Whether the items query is expected to be distinct.

        items_select_related (set of str, optional):
            Any models that are select-related for this query.

    Returns:
        list of dict:
        The list of expected queries.
    """
    items_q_tables = items_q_result['tables']
    items_q_join_types = items_q_result.get('join_types', {})
    items_q_num_joins = len(items_q_tables) - 1
    items_q_subqueries = items_q_result.get('subqueries', [])

    equeries = items_q_result.get('prep_equeries', [])

    if items_distinct:
        equeries += [
            {
                '__note__': 'Fetch the total number of items',
                'annotations': {
                    '__count': Count('*'),
                },
                'model': model,

                'subqueries': [
                    {
                        'distinct': True,
                        'join_types': items_q_join_types,
                        'model': model,
                        'num_joins': items_q_num_joins,
                        'subquery': True,
                        'tables': items_q_tables,
                        'where': items_q_result['q'],
                    },
                ],
            },
        ]
    else:
        equeries += [
            {
                '__note__': 'Fetch the total number of items',
                'annotations': {
                    '__count': Count('*'),
                },
                'join_types': items_q_join_types,
                'model': model,
                'num_joins': items_q_num_joins,
                'subqueries': items_q_subqueries,
                'tables': items_q_tables,
                'where': items_q_result['q'],
            },
        ]

    equeries += [
        {
            '__note__': 'Fetch a page of items',
            'distinct': items_distinct,
            'join_types': items_q_join_types,
            'limit': 25,
            'model': model,
            'num_joins': items_q_num_joins,
            'select_related': items_select_related,
            'subqueries': items_q_subqueries,
            'tables': items_q_tables,
            'where': items_q_result['q'],
        },
    ]

    return equeries



"""Unit tests for the VersionOne hosting service."""

from __future__ import annotations

from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.hostingsvcs.versionone import VersionOne


class VersionOneTests(HostingServiceTestCase[VersionOne]):
    """Unit tests for the VersionOne hosting service."""

    service_name = 'versionone'
    fixtures = ['test_scmtools']

    def test_service_support(self) -> None:
        """Testing VersionOne service support capabilities"""
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertFalse(self.service_class.supports_repositories)

    def test_get_bug_tracker_field(self) -> None:
        """Testing VersionOne.get_bug_tracker_field"""
        self.assertFalse(
            self.service_class.get_bug_tracker_requires_username())
        self.assertEqual(
            self.service_class.get_bug_tracker_field(None, {
                'versionone_url': 'http://versionone.example.com',
            }),
            'http://versionone.example.com/assetdetail.v1?Number=%s')



"""Sphinx extension for the Review Board documentation.

This provides general cross-reference roles/directives to help link related
content in the documentation.

Version Added:
    5.0.5
"""


def setup(app):
    app.add_crossref_type(
        directivename='rb-management-command',
        rolename='rb-management-command',
        indextemplate=('pair: %s; management command'),
    )

    # Django references.
    app.add_crossref_type(
        directivename='templatetag',
        rolename='ttag',
        indextemplate='pair: %s; template tag',
    )
